import os

import cv2
import h5py
import numpy as np
import seaborn as sns


class MultiClassOverlay(object):
    """Class to simplify overlaying images and labels."""

    def __init__(self, num_classes: int, color_palette=None, background_label=0, opacity=0.7):
        """
        Args:
            num_classes (int): Number of classes
            color_palette (optional): RGB tuples to use for category colors.
                Defaults to `sns.color_palette("bright")`.
            background_label (:obj:`int`, optional): Label to exclude for
                background.
            opacity (:obj:`float`, optional): Overlay transparency (0-1).
        """
        if not color_palette:
            color_palette = sns.color_palette("bright")
        effective_num_classes = num_classes - 1 if background_label is not None else num_classes
        if len(color_palette) < effective_num_classes:
            raise ValueError("Must provide at least %d colors" % effective_num_classes)

        if opacity < 0 or opacity > 1:
            raise ValueError("opacity must be between 0-1")

        self.num_classes = num_classes
        self.background_label = background_label
        self.opacity = opacity

        # set colormap
        color_palette = color_palette
        colormap = dict()
        cp_ind = 0
        for i in range(num_classes):
            if i == background_label:
                continue
            cp_ind += 1
            colormap[i] = color_palette[cp_ind]
        self.colormap = colormap

    def overlay(self, dirpath, volume: np.ndarray, logits: np.ndarray):
        # os.makedirs(dirpath, exist_ok=True)
        assert volume.ndim == 3, "Volume must be 3D array with shape HxWxD"
        assert logits.ndim == 4, "Labels must be 4D array with shape HxWxDxC"

        # labels are argmax(logits) in the class dimension
        labels = np.argmax(logits, axis=-1)
        labels_colored = self._apply_colormap(labels)

        for z in range(volume.shape[-1]):
            x_im = volume[..., z]
            label_overlay = labels_colored[..., z, :]

            # slice_name = "%03d.png" % (z + 1)
            # filepath = os.path.join(dirpath, slice_name)
            filepath = None
            self._im_overlay(x_im, label_overlay, filepath)

    def _apply_colormap(self, labels: np.ndarray):
        colormap = self.colormap
        background_label = self.background_label

        labels_colored = np.zeros(labels.shape + (3,))

        for c in np.unique(labels):
            if c == background_label:
                continue

            labels_colored[labels == c, :] = colormap[c]

        return (labels_colored * 255).astype(np.uint8)

    def _im_overlay(self, x, c_label, filepath=None):
        x_o = scale_img(np.squeeze(x))
        x_rgb = np.stack([x_o, x_o, x_o], axis=-1).astype(np.uint8)
        overlap_img = cv2.addWeighted(x_rgb, 1, c_label, self.opacity, 0)

        if filepath:
            cv2.imwrite(filepath, overlap_img)


class MultiClassOverlayNew(object):
    """
    Class to simplify overlaying images and labels.
    """

    def __init__(self, num_classes, color_palette=None, background_label=0, opacity=0.7):
        """
        Args:
            num_classes: Number of classes.
            color_palette: list of RGB tuples to use for color.
                Default seaborn.color_palette('bright').
            background_label: Label to exclude for background. Default: 0.
                                 To include background, set to None.
            opacity: How transparent overlay should be (0-1). Default: 0.7
        """
        if color_palette is None:
            color_palette = sns.color_palette("bright")
        effective_num_classes = num_classes - 1 if background_label is not None else num_classes
        if len(color_palette) < effective_num_classes:
            raise ValueError("Must provide at least %d colors" % effective_num_classes)

        if opacity < 0 or opacity > 1:
            raise ValueError("opacity must be between 0-1")

        self.num_classes = num_classes
        self.background_label = background_label
        self.opacity = opacity

        # set colormap
        color_palette = color_palette
        colormap = dict()
        cp_ind = 0
        for i in range(num_classes):
            if i == background_label:
                continue
            cp_ind += 1
            colormap[i] = color_palette[cp_ind]
        self.colormap = colormap

    def im_overlay(self, volume: np.ndarray, logits: np.ndarray, dirpath=None):
        """
        Overlay volume with labels.
        :param volume:
        :param logits:
        :param dirpath:
        :return:
        """
        if volume.ndim != 3:
            raise ValueError("Volume must be 3D array with shape [Y, X, Z].")
        if logits.ndim != 4:
            raise ValueError("Logits must be 4D binary array with shape [Y, X, Z, classes].")

        # Labels are argmax(logits) in the class dimension.
        labels = self.__logits_to_labels(logits)
        labels_colored = self.__apply_colormap(labels)

        if dirpath is not None:
            os.makedirs(dirpath, exist_ok=True)

        vol_rgb = np.zeros(volume.shape + (3,))
        for z in range(volume.shape[-1]):
            x_im = volume[..., z]
            label_overlay = labels_colored[..., z, :]

            slice_name = "%03d.png" % (z + 1)

            filepath = os.path.join(dirpath, slice_name) if dirpath is not None else None
            im_rgb = self.__im_overlay(x_im, label_overlay, filepath)
            vol_rgb[..., z, :] = im_rgb

        return vol_rgb

    def __logits_to_labels(self, logits: np.ndarray):
        assert logits.ndim == 4, "Logits must be 4D binary array with shape [Y, X, Z, classes]"
        logits = np.array(logits)
        for i in range(0, logits.shape[-1]):
            logits[..., i] *= i + 1
        return np.max(logits, axis=-1)

    def __apply_colormap(self, labels: np.ndarray):
        colormap = self.colormap
        background_label = self.background_label

        labels_colored = np.zeros(labels.shape + (3,))

        for c in np.unique(labels):
            if c == background_label:
                continue

            labels_colored[labels == c, :] = colormap[c]

        return (labels_colored * 255).astype(np.uint8)

    def __im_overlay(self, x, c_label, filepath=None):
        x_o = scale_img(np.squeeze(x))
        x_rgb = np.stack([x_o, x_o, x_o], axis=-1).astype(np.uint8)
        overlap_img = cv2.addWeighted(x_rgb, 1, c_label, self.opacity, 0)

        if filepath:
            cv2.imwrite(filepath, overlap_img)

        return overlap_img.astype(np.uint8)


def write_im_overlay(dir_path, xs, im_overlay, opacity=0.7):
    """
    Overlap input (xs) with mask (im_overlap) and save to directory
    :param dir_path: path to directory to save images
    :param xs: inputs
    :param im_overlay: overlay images
    """
    os.makedirs(dir_path, exist_ok=True)
    num_slices = xs.shape[0]
    for i in range(num_slices):
        x = scale_img(np.squeeze(xs[i, ...]))
        x = np.stack([x, x, x], axis=-1).astype(np.uint8)
        im = im_overlay[i]

        slice_name = "%03d.png" % i

        overlap_img = cv2.addWeighted(x, 1, im, opacity, 0)
        cv2.imwrite(os.path.join(dir_path, slice_name), overlap_img)


def write_sep_im_overlay(dir_path, xs, y_true, y_pred):
    """
    Overlap input (xs) with mask (im_overlap) and save to directory
    :param dir_path: path to directory to save images
    :param xs: inputs
    :param overlay: overlay images
    """
    correct_dir_path = os.path.join(dir_path, "true_pos")
    error_dir_path = os.path.join(dir_path, "error")
    os.makedirs(correct_dir_path, exist_ok=True)
    os.makedirs(error_dir_path, exist_ok=False)
    num_slices = xs.shape[0]
    for i in range(num_slices):
        x = scale_img(np.squeeze(xs[i, ...]))
        x = np.stack([x, x, x], axis=-1).astype(np.uint8)
        im_correct, im_error = generate_sep_ovlp_image(y_true[i, ...], y_pred[i, ...])

        slice_name = "%03d.png" % i

        overlap_img_correct = cv2.addWeighted(x, 1, im_correct, 1.0, 0)
        cv2.imwrite(os.path.join(correct_dir_path, slice_name), overlap_img_correct)

        overlap_img_error = cv2.addWeighted(x, 1, im_error, 1.0, 0)
        cv2.imwrite(os.path.join(error_dir_path, slice_name), overlap_img_error)


def generate_sep_ovlp_image(y_true, y_pred):
    """
    TODO: write comment
    :param y_true: numpy array of ground truth labels
    :param y_pred: numpy array of predicted labels
    :return: a BGR image
    """
    assert y_true.shape == y_pred.shape
    assert len(y_true.shape) == 2, "shape should be 2d, but is " + y_true.shape

    y_true = y_true.astype(np.bool)
    y_pred = y_pred.astype(np.bool)

    TP = y_true * y_pred
    FN = y_true * (~y_pred)
    FP = (~y_true) * y_pred

    # BGR format
    img_corr = (
        np.stack([np.zeros(TP.shape), TP, np.zeros(TP.shape)], axis=-1).astype(np.uint8) * 255
    )
    img_err = np.stack([FP, np.zeros(TP.shape), FN], axis=-1).astype(np.uint8) * 255

    return (img_corr, img_err)


def write_ovlp_masks(dir_path, y_true, y_pred):
    """
    Overlap ground truth with prediction and save to directory
    Red - false negative
    Green - true positive
    Blue - false negative
    :param dir_path: path to directory
    :param y_true: numpy array of ground truth labels
    :param y_pred: numpy array of predicted labels
    """
    os.makedirs(dir_path, exist_ok=True)
    y_true = np.squeeze(y_true)
    y_pred = np.squeeze(y_pred)

    assert y_true.shape == y_pred.shape
    ims = []
    num_slices = y_true.shape[0]
    for i in range(num_slices):
        slice_true = y_true[i, :, :]
        slice_pred = y_pred[i, :, :]
        img = generate_ovlp_image(slice_true, slice_pred)
        ims.append(img)

        slice_name = "%03d.png" % i
        cv2.imwrite(os.path.join(dir_path, slice_name), img)

    return ims


def write_mask(dir_path, y_true):
    """
    Save ground truth mask to directory
    :param dir_path: path to directory
    :param y_true: numpy array of ground truth labels
    """
    os.makedirs(dir_path, exist_ok=True)
    y_true = np.squeeze(y_true) * 255
    num_slices = y_true.shape[0]
    for i in range(num_slices):
        slice_name = "%03d.png" % i
        cv2.imwrite(os.path.join(dir_path, slice_name), y_true[i, :, :])


def write_prob_map(dir_path, y_probs):
    """
    Write probablity map for prediction as image (colormap jet)
    :param dir_path: path to directory
    :param y_probs: numpy array of prediction probabilities
    """
    os.makedirs(dir_path, exist_ok=True)
    y_probs = np.squeeze(y_probs)
    num_slices = y_probs.shape[0]
    for i in range(num_slices):
        slice_name = "%03d.png" % i
        im = y_probs[i, :, :] * 255
        im = im[..., np.newaxis].astype(np.uint8)
        imC = cv2.applyColorMap(im, cv2.COLORMAP_HOT)
        cv2.imwrite(os.path.join(dir_path, slice_name), imC)


def scale_img(im, scale=255):
    """
    Scale image from 0-scale
    :param im: input image
    :param scale: max value
    :return:
    """
    im = im.astype(np.float32)
    im = im - np.min(im)
    im = im / np.max(im)
    im *= scale

    return im


def generate_ovlp_image(y_true, y_pred):
    """
    Overlap ground truth and predicted labels
    :param y_true: numpy array of ground truth labels
    :param y_pred: numpy array of predicted labels
    :return: a BGR image
    """
    assert y_true.shape == y_pred.shape
    assert len(y_true.shape) == 2, "shape should be 2d, but is " + y_true.shape

    y_true = y_true.astype(np.bool)
    y_pred = y_pred.astype(np.bool)

    TP = y_true * y_pred
    FN = y_true * (~y_pred)
    FP = (~y_true) * y_pred

    # BGR format
    img = np.stack([FP, TP, FN], axis=-1).astype(np.uint8)
    img *= 255
    return img


def save_ims(output_dir: str):
    im_path = "%s.im" % output_dir
    with h5py.File(im_path, "r") as f:
        im = f["data"][:]

    seg_path = "%s.seg" % output_dir
    with h5py.File(seg_path, "r") as f:
        seg = f["data"][:].astype("float32")
        seg = seg[..., 0, 0]
    os.makedirs(output_dir, exist_ok=True)
    # save ims
    cv2.imwrite(os.path.join(output_dir, "im.png"), scale_img(im))

    # save segs
    cv2.imwrite(os.path.join(output_dir, "seg.png"), scale_img(seg))



from typing import Dict

import tensorflow as tf
from keras import backend as K
from keras.layers import Conv2D, Conv3D, ZeroPadding2D, ZeroPadding3D


def build_keras_config(class_name, **kwargs) -> Dict:
    """Builds config dictionary compatible with keras layer arguments."""
    return {"class_name": class_name, "config": kwargs}


def get_primary_shape(x: tf.Tensor):
    """Get sizes of the primary dimensions of x.

    Primary dimensions of a tensor are all dimensions that do not correspond to
    the batch dimension `B` or the channel dimension `C`.

    Args:
        x (tf.Tensor): Shape Bx(...)xC (channels_last) or BxCx(...) (channels_first).  # noqa

    Returns:
        list: primary dimensions.
    """
    x_shape = x.shape.as_list()
    x_shape = x_shape[1:-1] if K.image_data_format() == "channels_last" else x_shape[2:]

    return x_shape


def zero_pad_like(x: tf.Tensor, y: tf.Tensor, x_shape=None, y_shape=None):
    """Zero pads input (x) to size of target (y).

    Padding is symmetric when difference in dimension size is multiple of 2.
    Otherwise, the bottom padding is 1 larger than the top padding.
    Assumes channels are last dimension.

    Primary dimensions of a tensor are all dimensions that do not correspond to
    the batch dimension or the channel dimension.

    Args:
        x (tf.Tensor): Input tensor.
        y (tf.Tensor): Target tensor.
        x_shape (Sequence[int]): Expected shape of `x`. Required when primary
            `x` dimensions have sizes `None`.
        y_shape (Sequence[int]): Like `x_shape`, but for `y`.

    Returns:
        tf.Tensor: Zero-padded tensor
    """
    if not x_shape:
        x_shape = get_primary_shape(x)
    if not y_shape:
        y_shape = get_primary_shape(y)

    assert not any(s is None for s in x_shape)
    assert not any(s is None for s in y_shape)

    if x_shape == y_shape:
        return x
    diff = [y_s - x_s for x_s, y_s in zip(x_shape, y_shape)]
    assert all(d >= 0 for d in diff), "x must be smaller than y in all dimensions"

    if len(diff) == 2:
        padder = ZeroPadding2D
    elif len(diff) == 3:
        padder = ZeroPadding3D
    else:
        raise ValueError("Zero padding available for 2D or 3D images only")

    padding = [d // 2 if d % 2 == 0 else (d // 2, d // 2 + 1) for d in diff]
    x = padder(padding)(x)
    return x


def add_sem_seg_activation(
    x: tf.Tensor,
    num_classes: int,
    activation: str = "sigmoid",
    conv_type=None,
    kernel_initializer=None,
    seed=None,
) -> tf.Tensor:
    """Standardized output layer for semantic segmentation using 1x1 conv.

    Args:
        x (tf.Tensor): Input tensor.
        num_classes (int): Number of classes
        activation (str, optional): Activation type. Typically `'sigmoid'` or
            `'softmax'`.
        conv_type: Either `Conv2D` or `Conv3D`.
        kernel_initializer: Kernel initializer accepted by
            `Conv2D` or `Conv3D`.
        seed (int, optional): Kernel intialization seed. Ignored if
            `kernel_initializer` is a config dict.
    """

    # Initializing kernel weights to 1 and bias to 0.
    # i.e. without training, the x would be a sigmoid activation on each
    # pixel of the input
    if not conv_type:
        conv_type = Conv2D
    else:
        assert conv_type in [Conv2D, Conv3D]
    if not kernel_initializer:
        kernel_initializer = {"class_name": "glorot_uniform", "config": {"seed": seed}}
    elif isinstance(kernel_initializer, str):
        kernel_initializer = {"class_name": kernel_initializer, "config": {"seed": seed}}

    x = conv_type(
        num_classes,
        1,
        activation=activation,
        kernel_initializer=kernel_initializer,
        name="output_activation",
    )(x)
    return x



import importlib
import importlib.util
import logging
import os
import random
import sys
from datetime import datetime
from typing import Tuple

import numpy as np
import tensorflow as tf

__all__ = []

_ENV_SETUP_DONE = False
_SETTINGS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.settings"))
_TF_VERSION = None


def generate_seed():
    """Generate a random seed."""
    seed = os.getpid() + int(datetime.now().strftime("%S%f")) + int.from_bytes(os.urandom(2), "big")
    logger = logging.getLogger(__name__)
    logger.info("Generated random seed {}".format(seed))
    return seed


def seed_all_rng(seed=None):
    """Set random seed for RNG in tensorflow, numpy, and python.

    Args:
        seed (int, optional): The random seed. If not specified, this function
            generates a random seed using :func:`generate_seed`.
    """
    logger = logging.getLogger(__name__)
    if seed is None:
        seed = generate_seed()

    np.random.seed(seed)
    try:
        tf.random.set_seed(seed)
    except AttributeError:
        tf.random.set_random_seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    logger.info("Using random seed {}".format(seed))


def is_debug():
    return os.environ.get("MEDSEGPY_RUN_MODE", "") == "debug"


# from https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path  # noqa
def _import_file(module_name, file_path, make_importable=False):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if make_importable:
        sys.modules[module_name] = module
    return module


def _configure_libraries():
    """
    Configurations for some libraries.
    """
    # An environment option to disable `import cv2` globally,
    # in case it leads to negative performance impact
    disable_cv2 = int(os.environ.get("MEDSEGPY_DISABLE_CV2", False))
    if disable_cv2:
        sys.modules["cv2"] = None
    else:
        # Disable opencl in opencv since its interaction with cuda often
        # has negative effects
        # This envvar is supported after OpenCV 3.4.0
        os.environ["OPENCV_OPENCL_RUNTIME"] = "disabled"
        try:
            import cv2

            if int(cv2.__version__.split(".")[0]) >= 3:
                cv2.ocl.setUseOpenCL(False)
        except ImportError:
            pass


def setup_environment():
    """Perform environment setup work. The default setup is a no-op, but this
    function allows the user to specify a Python source file or a module in
    the $MEDSEGPY_ENV_MODULE environment variable, that performs
    custom setup work that may be necessary to their computing environment.
    """
    global _ENV_SETUP_DONE
    if _ENV_SETUP_DONE:
        return
    _ENV_SETUP_DONE = True

    _configure_libraries()

    custom_module_path = os.environ.get("MEDSEGPY_ENV_MODULE")

    if custom_module_path:
        setup_custom_environment(custom_module_path)
    else:
        # The default setup is a no-op
        pass


def setup_custom_environment(custom_module):
    """
    Load custom environment setup by importing a Python source file or a
    module, and run the setup function.
    """
    if custom_module.endswith(".py"):
        module = _import_file("medsegpy.utils.env.custom_module", custom_module)
    else:
        module = importlib.import_module(custom_module)
    assert hasattr(module, "setup_environment") and callable(module.setup_environment), (
        "Custom environment module defined in {} does not have the "
        "required callable attribute 'setup_environment'."
    ).format(custom_module)
    module.setup_environment()


def supports_wandb():
    return "wandb" in sys.modules and not is_debug()


def tf_version() -> Tuple[int, ...]:
    global _TF_VERSION
    if not _TF_VERSION:
        import tensorflow as tf

        _TF_VERSION = [int(x) for x in tf.__version__.split(".")[:2]]
    return tuple(_TF_VERSION)


def is_tf2():
    """Returns `True` if running tensorflow 2.X"""
    version = tf_version()
    return version[0] == 2


def settings_dir():
    return os.environ.get("MEDSEGPY_SETTINGS", _SETTINGS_DIR)



"""Learning rate schedulers.

Usage:
    >>> callbacks = []  # list of callbacks to be used sith `fit_generator`
    >>> scheduler = step_decay(...)
    >>> callbacks.append(keras.callback.LearningRateScheduler(scheduler))
"""


def step_decay(initial_lr, min_lr, drop_factor, drop_rate):
    """Learning rate drops by factor of `drop_factor` every `drop_rate` epochs.

    For legacy purposes, the first drop occurs after `drop_rate - 1` epochs.
    For example, if `drop_rate = 3`, the first decay will occur after 2 epochs.
    Subsequently, the learning rate will drop every 3 epochs.

    Args:
        initial_lr: initial learning rate (default = 1e-4)
        min_lr: minimum learning rate (default = None)
        drop_factor: factor to drop (default = 0.8)
        drop_rate: rate of learning rate drop (default = 1.0 epochs)

    Returns:
        func: To be used with :class`keras.callbacks.LearningRateScheduler`
    """
    initial_lr = initial_lr
    drop_factor = drop_factor
    drop_rate = drop_rate
    min_lr = min_lr

    def callback(epoch):
        import math

        lrate = initial_lr * math.pow(drop_factor, math.floor((1 + epoch) / drop_rate))
        if lrate < min_lr:
            lrate = min_lr

        return lrate

    return callback



import ast
import configparser
import copy
import logging
import os
import warnings
from itertools import groupby
from typing import Any, Dict, Tuple

import yaml
from fvcore.common.file_io import PathManager

from medsegpy.cross_validation import cv_util
from medsegpy.losses import DICE_LOSS, L2_LOSS, get_training_loss_from_str
from medsegpy.utils import utils as utils

logger = logging.getLogger(__name__)

# Keys that have been deprecated.
DEPRECATED_KEYS = [
    "NUM_CLASSES",
    "TRAIN_FILES_CV",
    "VALID_FILES_CV",
    "TEST_FILES_CV",
    "USE_STEP_DECAY",
    "PIK_SAVE_PATH_DIR",
    "PIK_SAVE_PATH",
    "TF_LOG_DIR",
    "TRAIN_PATH",
    "VALID_PATH",
    "TEST_PATH",
    "PLOT_MODEL_PATH",
    "FINE_TUNE",
    "LEARN_FILES",
    "DEBUG",
    "TEST_RESULT_PATH",
    "TEST_RESULTS_FOLDER_NAME",
    "DATE_TIME_STR",
]

RENAMED_KEYS = {
    "CP_SAVE_PATH": "OUTPUT_DIR",
    "CP_SAVE_TAG": "MODEL_NAME",
    "INIT_WEIGHT_PATH": "INIT_WEIGHTS",
    "TISSUES": "CATEGORIES",
}

BASE_KEY = "_BASE_"


class Config(object):
    """A config object that is 1-to-1 with supported models.

    Each subclass of :class:`Config` corresponds to a specific model
    architecture.
    """

    VERSION = 7

    # Model name specific to config. Cannot be changed.
    MODEL_NAME = ""

    # Experiment name and description of the config.
    EXP_NAME = ""
    DESCRIPTION = ""

    # Learning type. Possible Values:
    # 1) "self-supervised" -- self-supervised learning
    # 2) "supervised" -- supervised learning
    LEARNING_TAG = ""

    # The current task for which the model is trained
    # -- Possible Values: "inpainting", "segmentation"
    PRIMARY_TASK = ""

    # The layers corresponding to the pretrained portion of a
    # model during self-supervised learning.
    SS_LOADED_LAYERS = []

    # Loss function in form (id, output_mode)
    LOSS = DICE_LOSS
    CLASS_WEIGHTS = None
    # Class name for robust loss computation
    ROBUST_LOSS_NAME = ""
    ROBUST_LOSS_STEP_SIZE = 1e-1

    # PIDS to include, None = all pids
    PIDS = None

    # Training and validation image size
    IMG_SIZE = (288, 288, 1)
    # Patch-based image loading params used with PatchDataLoader
    IMG_PAD_SIZE = ()
    IMG_PAD_MODE = "edge"
    IMG_STRIDE = ()

    # Training parameters
    N_EPOCHS = 100
    AUGMENT_DATA = False

    # Step Decay params
    INITIAL_LEARNING_RATE = 1e-4
    LR_SCHEDULER_NAME = ""
    MIN_LEARNING_RATE = 1e-8
    DROP_FACTOR = 0.7
    DROP_RATE = 1.0
    LR_MIN_DELTA = 1e-4
    LR_PATIENCE = 0
    LR_COOLDOWN = 0
    NUM_GRAD_STEPS = 1

    # ADAM optimizer decay
    ADAM_DECAY = 0.0
    USE_AMSGRAD = False

    # Early stopping criterion
    USE_EARLY_STOPPING = False
    EARLY_STOPPING_MIN_DELTA = 0.0
    EARLY_STOPPING_PATIENCE = 0
    EARLY_STOPPING_CRITERION = "val_loss"

    # Batch sizes
    TRAIN_BATCH_SIZE = 12
    VALID_BATCH_SIZE = 35
    TEST_BATCH_SIZE = 72

    # Categories
    CATEGORIES = []
    INCLUDE_BACKGROUND = False

    # File Types
    FILE_TYPES = ["im"]

    # A dictionary specifying which pretrained weights to load
    #
    # Example:
    #   {
    #       "path": "",
    #       "weights":
    #           [
    #               {
    #                   "include_words": ["decoder"],
    #                   "exclude_words": [],
    #                   "slice_indices": [0, "until"]
    #               }
    #           ]
    #   }
    #
    # The above example will load the weights for all layers from the
    # first layer until the layer right before the first layer
    # with the word "decoder" in its name
    PRETRAINED_WEIGHTS_PATH = {}

    # The path to the config file of the pretrained model
    PRETRAINED_CONFIG_PATH = ""

    # Boolean determining whether or not the pretrained weights
    # specified in PRETRAINED_WEIGHTS_PATH will be frozen
    # during training
    FREEZE_PRETRAINED = False

    INIT_WEIGHTS = ""
    FREEZE_LAYERS = ()

    # Dataset names
    TRAIN_DATASET = ""
    VAL_DATASET = ""
    TEST_DATASET = ""

    # Cross-Validation-Parameters
    USE_CROSS_VALIDATION = False
    CV_FILE = ""
    CV_K = 0
    CV_TRAIN_BINS = []
    CV_VALID_BINS = []
    CV_TEST_BINS = []
    __CV_TRAIN_FILES__ = None
    __CV_VALID_FILES__ = None
    __CV_TEST_FILES__ = None

    # Training Model Paths
    OUTPUT_DIR = ""

    # DataLoader tag.
    TAG = "DefaultDataLoader"
    PRELOAD_DATA = False

    # Type of normalization layer
    NORMALIZATION = "BatchNorm"
    # Arguments for the normalization layer
    NORMALIZATION_ARGS = {"axis": -1, "momentum": 0.95, "epsilon": 0.001}
    # Boolean specifying if weight standardization should be used
    WEIGHT_STANDARDIZATION = False

    # Weights kernel initializer.
    KERNEL_INITIALIZER = "he_normal"

    # System params
    NUM_WORKERS = 1
    SEED = None

    # Evaluation params
    TEST_WEIGHT_PATH = ""
    TEST_METRICS = ["DSC", "VOE", "ASSD", "CV"]

    # PREPROCESSING: a list of MedTransform or TransformGen class names
    #                   (look at medsegpy/data/transforms/transform_gen.py")
    # PREPROCESSING_ARGS: a list of dictionaries, where each dictionary
    #                       contains the value of the __init__ arguments for
    #                       the corresponding class listed in PREPROCESSING
    PREPROCESSING = []
    PREPROCESSING_ARGS = []

    def __init__(self, cp_save_tag, state="training", create_dirs=True):
        if state not in ["testing", "training"]:
            raise ValueError("state must either be 'training' or 'testing'")

        self.MODEL_NAME = cp_save_tag
        self.STATE = state

    def init_cross_validation(
        self,
        train_files,
        valid_files,
        test_files,
        train_bins,
        valid_bins,
        test_bins,
        cv_k,
        cv_file,
        output_dir,
    ):
        """Initialize config for cross validation.

        Returns:
            Config: A deep copy of the config. This copy is initialized for
                cross validation.
        """
        assert (
            self.STATE == "training"
        ), "Initializing cross-validation must be done in training state"

        config = copy.deepcopy(self)
        config.USE_CROSS_VALIDATION = True
        config.CV_TRAIN_BINS = train_bins
        config.CV_VALID_BINS = valid_bins
        config.CV_TEST_BINS = test_bins
        config.CV_K = cv_k
        config.CV_FILE = cv_file

        config.__CV_TRAIN_FILES__ = train_files
        config.__CV_VALID_FILES__ = valid_files
        config.__CV_TEST_FILES__ = test_files

        os.makedirs(output_dir, exist_ok=True)
        config.OUTPUT_DIR = output_dir

        return config

    def save_config(self):
        """Save params of config to ini file."""
        filepath = os.path.join(self.OUTPUT_DIR, "config.yaml")
        config_vars = self.to_dict()

        # Save config
        with PathManager.open(filepath, "w") as configfile:
            yaml.safe_dump(config_vars, configfile)

        logger.info("Full config saved to {}".format(os.path.abspath(filepath)))

    def to_dict(self):
        """Returns a dictionary representation of the config.

        This representation only contains public instance variable fields.
        Private and protected instance variables are excluded. Methods are excluded.

        Returns:
            dict: The config as a dictionary.
        """
        members = [
            attr
            for attr in dir(self)
            if not attr.startswith("__")
            and not callable(getattr(self, attr))
            and not (hasattr(type(self), attr) and isinstance(getattr(type(self), attr), property))
        ]
        return {m_var: getattr(self, m_var) for m_var in members}

    def _parse_special_attributes(self, full_key: str, value: Any) -> Tuple[str, Any]:
        """Special parsing values for attributes.

        Used when loading config from a file or from list.

        Args:
            full_key (str): Upper case attribute representation.
            value (Any): Corresponding value.
        """
        # Avoid circular dependencies.
        from medsegpy.data import MetadataCatalog  # noqa
        from medsegpy.data.data_loader import LEGACY_DATA_LOADER_NAMES  # noqa
        from medsegpy.modeling.meta_arch.build import LEGACY_MODEL_NAMES  # noqa

        prev_key, prev_val = full_key, value

        if full_key in ("TRAIN_PATH", "VALID_PATH", "TEST_PATH") and value:
            mapping = {
                "TRAIN_PATH": "TRAIN_DATASET",
                "VALID_PATH": "VAL_DATASET",
                "TEST_PATH": "TEST_DATASET",
            }
            value = MetadataCatalog.convert_path_to_dataset(value)
            full_key = mapping[full_key]
            logger.info("Converting {} -> {}: {} -> {}".format(prev_key, full_key, prev_val, value))
        elif full_key == "TAG" and value in LEGACY_DATA_LOADER_NAMES:
            value = LEGACY_DATA_LOADER_NAMES[value]
            logger.info("Converting {}: {} -> {}".format(full_key, prev_val, value))
        elif full_key == "MODEL_NAME" and value in LEGACY_MODEL_NAMES:
            value = LEGACY_MODEL_NAMES[value]
            logger.info("Converting {}: {} -> {}".format(full_key, prev_val, value))
        elif full_key == "LOSS" and isinstance(value, str):
            try:
                value = get_training_loss_from_str(value)
            except ValueError:
                pass
        elif full_key == "OUTPUT_DIR":
            value = PathManager.get_local_path(value)

        return full_key, value

    def merge_from_file(self, cfg_filename: str, new_allowed=False):
        """Load a ini or yaml config file and merge it with this object.

        "MODEL_NAME" must be specified in the file.

        Args:
            cfg_filename (str): File path to yaml or ini file.
            new_allowed (bool, optional): If `True`, new keys, not already defined,
                are allowed.
        """
        # Avoid circular dependencies.
        from medsegpy.modeling.meta_arch.build import LEGACY_MODEL_NAMES

        vars_dict = self._load_dict_from_file(cfg_filename)

        # Load information from base files first.
        if BASE_KEY in vars_dict:
            vars_dict = self.parse_base_config(cfg_filename)

        # TODO: Handle cp save tag as a protected key.
        model_name = (
            vars_dict["MODEL_NAME"] if "MODEL_NAME" in vars_dict else vars_dict["CP_SAVE_TAG"]
        )
        if model_name in LEGACY_MODEL_NAMES:
            model_name = LEGACY_MODEL_NAMES[model_name]
        if model_name != self.MODEL_NAME:
            raise ValueError(
                "Wrong config. Expected {}. Got {}".format(self.MODEL_NAME, model_name)
            )

        for full_key, value in vars_dict.items():
            full_key = str(full_key).upper()

            if full_key in RENAMED_KEYS:
                new_name = RENAMED_KEYS[full_key]
                logger.warning("Key {} has been renamed to {}".format(full_key, new_name))
                full_key = new_name

            full_key, value = self._parse_special_attributes(full_key, value)

            if full_key in DEPRECATED_KEYS:
                logger.warning("Key {} is deprecated, not loading".format(full_key))
                continue

            if not new_allowed and not hasattr(self, full_key):
                raise ValueError("Key {} does not exist.".format(full_key))

            is_new = not hasattr(self, full_key)

            if is_new or full_key == "TRAIN_DATASET":
                value = self._decode_cfg_value(value, "auto")
            else:
                value = self._decode_cfg_value(value, type(self.__getattribute__(full_key)))
                value = _check_and_coerce_cfg_value_type(
                    value, self.__getattribute__(full_key), full_key
                )

            # Loading config
            self.__setattr__(full_key, value)

    def merge_from_list(self, cfg_list):
        """Merge config (keys, values) in a list (e.g. from command line).

        For example, cfg_list = ['FOO_BAR', 0.5, 'BAR_FOO', (0,3,4)]
        """
        _error_with_logging(
            len(cfg_list) % 2 == 0,
            "Override list has odd length: {}; " "it must be a list of pairs".format(cfg_list),
        )

        for full_key, v in zip(cfg_list[0::2], cfg_list[1::2]):
            if full_key == "MODEL_NAME":
                raise ValueError("Cannot change key MODEL_NAME")
            if self.key_is_deprecated(full_key):
                continue

            if self.key_is_renamed(full_key):
                self.raise_key_rename_error(full_key)

            _error_with_logging(
                hasattr(self, full_key),
                "Non-existent key: {}".format(full_key),
                error_type=KeyError,
            )
            value = self._decode_cfg_value(v, type(self.__getattribute__(full_key)))
            value = _check_and_coerce_cfg_value_type(
                value, self.__getattribute__(full_key), full_key
            )
            self.__setattr__(full_key, value)

    @classmethod
    def _decode_cfg_value(cls, value, data_type):
        """
        Decodes a raw config value (e.g., from a yaml config files or command
        line argument) into a Python object.

        If the value is a dict, it will be interpreted as a new CfgNode.
        If the value is a str, it will be evaluated as literals.
        Otherwise it is returned as-is.
        """
        # Configs parsed from raw yaml will contain dictionary keys that need to
        # be converted to CfgNode objects
        """
        Convert string to relevant data type
        :param var_string: variable as a string (e.g.: '[0]', '1', 'hellow')
        :param data_type: the type of the data
        :return: string converted to data_type
        """
        if isinstance(value, dict):
            return {k: cls._decode_cfg_value(v, "auto") for k, v in value.items()}
        elif not isinstance(value, str):
            return value

        if data_type is str:
            return str(value)
        elif data_type is float:
            return float(value)
        elif data_type is int:
            return int(value)
        else:
            try:
                return ast.literal_eval(value)
            except (SyntaxError, ValueError) as e:
                if data_type == "auto":
                    return value
                else:
                    raise e

    def key_is_deprecated(self, full_key):
        """Test if a key is deprecated."""
        if full_key in DEPRECATED_KEYS:
            logger.warning("Deprecated config key (ignoring): {}".format(full_key))
            return True
        return False

    def key_is_renamed(self, full_key):
        """Test if a key is renamed."""
        return full_key in RENAMED_KEYS

    def add_new_key(self, key, value) -> None:
        """Add a new key to the config schema.

        Args:
            key (str): Key name
            value (Any): Value of the key to set.
        """
        if key in self.__dict__:
            raise KeyError("Key {} already exists in config.".format(key))
        value = self._decode_cfg_value(value, "auto")
        self.__setattr__(key, value)

    def raise_key_rename_error(self, full_key):
        new_key = RENAMED_KEYS[full_key]
        if isinstance(new_key, tuple):
            msg = " Note: " + new_key[1]
            new_key = new_key[0]
        else:
            msg = ""
        raise KeyError(
            "Key {} was renamed to {}; please update your config.{}".format(full_key, new_key, msg)
        )

    @classmethod
    def parse_base_config(cls, cfg_filename: str) -> Dict:
        local_dir = os.path.dirname(PathManager.get_local_path(cfg_filename))
        vars_dict = cls._load_dict_from_file(cfg_filename)
        base_cfg_filename = vars_dict.pop(BASE_KEY, None)
        if base_cfg_filename:
            if base_cfg_filename.startswith("~"):
                base_cfg_filename = os.path.expanduser(base_cfg_filename)
            elif any(map(base_cfg_filename.startswith, ["/", "https://", "http://"])):
                raise ValueError("Remote configs not currently supported.")
            else:
                # the path to base cfg is relative to the config file itself.
                base_cfg_filename = os.path.join(local_dir, base_cfg_filename)
            base_dict = cls.parse_base_config(base_cfg_filename)
            base_dict.update(vars_dict)
            vars_dict = base_dict

        assert BASE_KEY not in vars_dict, "{} should be popped off!".format(BASE_KEY)
        return vars_dict

    @classmethod
    def _load_dict_from_file(cls, cfg_filename):
        filename = PathManager.get_local_path(cfg_filename)
        if filename.endswith(".ini"):
            cfg = configparser.ConfigParser()
            if not os.path.isfile(filename):
                raise FileNotFoundError("Config file {} not found".format(filename))
            cfg.read(filename)
            vars_dict = cfg["DEFAULT"]
            vars_dict = {k.upper(): v for k, v in vars_dict.items()}
        elif filename.endswith(".yaml") or filename.endswith(".yml"):
            with open(filename, "r") as f:
                vars_dict = yaml.safe_load(f)
        else:
            raise ValueError("file {} not supported".format(filename))

        return vars_dict

    def set_attr(self, attr, val):
        """
        Wrapper method to set attributes of config
        :param attr: a string attr
        :param val: value to set attribute to

        :raises ValueError: if attr is not a string
                            if attr does not exist for the config
                            if type of val is different from the default type
        """
        if type(attr) is not str:
            raise ValueError("attr must be of type str")

        if not hasattr(self, attr):
            raise ValueError("The attribute %s does not exist" % attr)
        curr_val = self.__getattribute__(attr)

        if type(val) is str and type(curr_val) is not str:
            val = utils.convert_data_type(var_string=val, original=curr_val)

        if curr_val is not None and (type(val) != type(curr_val)):
            raise ValueError(
                "%s is of type %s. Expected %s" % (attr, str(type(val)), str(type(curr_val)))
            )

        self.__setattr__(attr, val)

    def __getitem__(self, key):
        if not hasattr(self, key):
            raise KeyError(f"Key {key} does not exist")
        return getattr(self, key)

    def get(self, key, default=None):
        keys = key.split(".")
        try:
            val = self
            for k in keys:
                val = val[k]
        except KeyError:
            return default

        return val

    def change_to_test(self):
        """
        Initialize testing state
        """
        self.STATE = "testing"

        # if cross validation is enabled, load testing cross validation bin
        if self.USE_CROSS_VALIDATION:
            assert self.CV_FILE, "No cross-validation file found in config"
            cv_processor = cv_util.CrossValidationProcessor(self.CV_FILE)
            bins = (self.CV_TRAIN_BINS, self.CV_VALID_BINS, self.CV_TEST_BINS)

            train_files, valid_files, test_files = cv_processor.get_fnames(bins)

            self.__CV_TRAIN_FILES__ = train_files
            self.__CV_VALID_FILES__ = valid_files
            self.__CV_TEST_FILES__ = test_files

    def summary(self, additional_vars=None):
        """
        Print config summary
        :param additional_vars: additional list of variables to print
        :return:
        """

        summary_vals = ["MODEL_NAME", "TAG", ""]
        summary_vals.extend(
            [
                "TRAIN_DATASET",
                "VAL_DATASET",
                "TEST_DATASET",
                "",
                "CATEGORIES",
                "",
                "IMG_SIZE",
                "",
                "N_EPOCHS",
                "AUGMENT_DATA",
                "LOSS",
                "CLASS_WEIGHTS",
                "ROBUST_LOSS_NAME"
                if self.ROBUST_LOSS_NAME
                else "" "ROBUST_LOSS_STEP_SIZE"
                if self.ROBUST_LOSS_NAME
                else "" "",
                "USE_CROSS_VALIDATION",
                "CV_K" if self.USE_CROSS_VALIDATION else "",
                "CV_FILE" if self.USE_CROSS_VALIDATION else "",
                "CV_TRAIN_BINS" if self.USE_CROSS_VALIDATION else "",
                "CV_VALID_BINS" if self.USE_CROSS_VALIDATION else "",
                "CV_TEST_BINS" if self.USE_CROSS_VALIDATION else "",
                "" "TRAIN_BATCH_SIZE",
                "VALID_BATCH_SIZE",
                "TEST_BATCH_SIZE",
                "",
                "NUM_GRAD_STEPS",
                "",
                "INITIAL_LEARNING_RATE",
                "LR_SCHEDULER_NAME",
                "DROP_FACTOR" if self.LR_SCHEDULER_NAME else "",
                "DROP_RATE" if self.LR_SCHEDULER_NAME else "",
                "MIN_LEARNING_RATE" if self.LR_SCHEDULER_NAME else "",
                "LR_MIN_DELTA" if self.LR_SCHEDULER_NAME else "",
                "LR_PATIENCE" if self.LR_SCHEDULER_NAME else "",
                "LR_COOLDOWN" if self.LR_SCHEDULER_NAME else "",
                "" "USE_EARLY_STOPPING",
                "EARLY_STOPPING_MIN_DELTA" if self.USE_EARLY_STOPPING else "",
                "EARLY_STOPPING_PATIENCE" if self.USE_EARLY_STOPPING else "",
                "EARLY_STOPPING_CRITERION" if self.USE_EARLY_STOPPING else "",
                "",
                "KERNEL_INITIALIZER",
                "SEED" if self.SEED else "",
                "" "INIT_WEIGHTS",
                "",
                "TEST_WEIGHT_PATH",
                "TEST_METRICS",
                "" "NUM_WORKERS",
                "OUTPUT_DIR",
                "",
            ]
        )
        if additional_vars:
            summary_vals.extend(additional_vars)

        # Remove consecutive elements in summary vals that are the same
        summary_vals = [x[0] for x in groupby(summary_vals)]

        logger.info("")
        logger.info("==" * 40)
        logger.info("Config Summary")
        logger.info("==" * 40)

        for attr in summary_vals:
            if attr == "":
                logger.info("")
                continue
            logger.info(attr + ": " + str(self.__getattribute__(attr)))

        logger.info("==" * 40)
        logger.info("")

    def get_num_classes(self):
        if self.INCLUDE_BACKGROUND:
            return len(self.CATEGORIES) + 1

        return len(self.CATEGORIES)

    def num_neighboring_slices(self):
        return None

    @property
    def testing(self):
        return self.STATE == "testing"

    @property
    def training(self):
        return self.STATE == "training"


class DeeplabV3Config(Config):
    """
    Configuration for 2D Deeplabv3+ architecture
    (https://arxiv.org/abs/1802.02611).
    """

    MODEL_NAME = "DeeplabV3Plus"

    OS = 16
    DIL_RATES = (2, 4, 6)
    AT_DIVISOR = 2
    DROPOUT_RATE = 0.1

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_attrs = ["OS", "DIL_RATES", "DROPOUT_RATE"]
        super().summary(summary_attrs)


class SegnetConfig(Config):
    """
    Configuration for 2D Segnet architecture (https://arxiv.org/abs/1505.07293)
    """

    MODEL_NAME = "segnet_2d"

    TRAIN_BATCH_SIZE = 15

    DEPTH = 6
    NUM_CONV_LAYERS = [2, 2, 3, 3, 3, 3]
    NUM_FILTERS = [64, 128, 256, 256, 512, 512]

    SINGLE_BN = False
    CONV_ACT_BN = False
    USE_BOTTLENECK = False

    INITIAL_LEARNING_RATE = 1e-3

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_attrs = ["DEPTH", "NUM_CONV_LAYERS", "NUM_FILTERS"]
        super().summary(summary_attrs)


class ContextEncoderConfig(Config):
    """
    Configuration for the context encoder.

    Reference:
        Pathak et al. "Context Encoders: Feature Learning by Inpainting". CVPR. 2016.
    """

    MODEL_NAME = "ContextEncoder"
    NUM_FILTERS = [[32, 32], [64, 64], [128, 128], [256, 256]]

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_vars = ["NUM_FILTERS"]
        if additional_vars:
            summary_vars.extend(additional_vars)
        super().summary(summary_vars)


class ContextUNetConfig(Config):
    """
    Configuration for the ContextUNet model.

    This model will incorporate the ContextEncoder model as well.

    Reference:
        Pathak et al. "Context Encoders: Feature Learning by Inpainting". CVPR. 2016.
    """

    MODEL_NAME = "ContextUNet"
    NUM_FILTERS = [[32, 32], [64, 64], [128, 128], [256, 256]]

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_vars = ["NUM_FILTERS"]
        if additional_vars:
            summary_vars.extend(additional_vars)
        super().summary(summary_vars)


class ContextInpaintingConfig(ContextUNetConfig):
    """
    Configuration for the ContextInpainting model.
    """

    MODEL_NAME = "ContextInpainting"
    LOSS = L2_LOSS

    # Define preprocessing transforms
    PREPROCESSING = ["CoarseDropout"]
    PREPROCESSING_ARGS = [
        {
            "max_holes": 25,
            "max_height": 50,
            "max_width": 50,
            "min_holes": 0,
            "min_height": 25,
            "min_width": 25,
            "fill_value": 0.257,
        }
    ]


class ContextSegmentationConfig(ContextUNetConfig):
    """
    Configuration for the ContextSegmentation model.
    """

    MODEL_NAME = "ContextSegmentation"


class UNetConfig(Config):
    """
    Configuration for 2D U-Net architecture (https://arxiv.org/abs/1505.04597)
    """

    MODEL_NAME = "UNet2D"

    INIT_UNET_2D = False

    INITIAL_LEARNING_RATE = 2e-2
    DROP_FACTOR = 0.8 ** (1 / 5)
    DROP_RATE = 1.0
    TRAIN_BATCH_SIZE = 35

    DEPTH = 6
    NUM_FILTERS = None

    # Boolean arguments for attention and deep supervision
    USE_ATTENTION = False
    USE_DEEP_SUPERVISION = False

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_vars = ["DEPTH", "NUM_FILTERS", ""]
        if additional_vars:
            summary_vars.extend(additional_vars)
        super().summary(summary_vars)


class ResidualUNet(Config):
    """
    Configuration for 2D Residual U-Net architecture
    """

    MODEL_NAME = "res_unet"

    DEPTH = 6
    NUM_FILTERS = None

    DROPOUT_RATE = 0.0
    LAYER_ORDER = ["relu", "bn", "dropout", "conv"]

    USE_SE_BLOCK = False
    SE_RATIO = 8

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_attrs = [
            "DEPTH",
            "NUM_FILTERS",
            "DROPOUT_RATE",
            "",
            "LAYER_ORDER",
            "",
            "USE_SE_BLOCK",
            "SE_RATIO",
        ]
        super().summary(summary_attrs)

    def num_neighboring_slices(self):
        return self.IMG_SIZE[-1] if self.IMG_SIZE[-1] != 1 else None


class UNet2_5DConfig(UNetConfig):
    """
    Configuration for 3D U-Net architecture
    """

    MODEL_NAME = "UNet2D"

    IMG_SIZE = (288, 288, 7)

    N_EPOCHS = 20
    AUGMENT_DATA = False
    INITIAL_LEARNING_RATE = 1e-2

    DROP_RATE = 1.0
    DROP_FACTOR = 0.8

    def __init__(self, state="training", create_dirs=True):
        warnings.warn("UNet2_5DConfig is deprecated. Use UNet2DConfig instead.", DeprecationWarning)
        super().__init__(state, create_dirs)

    def num_neighboring_slices(self):
        return self.IMG_SIZE[2]


class UNet3DConfig(UNetConfig):
    MODEL_NAME = "UNet3D"

    IMG_SIZE = (288, 288, 4, 1)

    N_EPOCHS = 20
    INITIAL_LEARNING_RATE = 1e-2

    DROP_RATE = 1.0
    DROP_FACTOR = 0.8

    TAG = "oai_3d"

    SLICE_SUBSET = None  # 1 indexed inclusive - i.e. (5, 64) means slices [5, 64]

    NUM_FILTERS = [32, 64, 128, 256, 512, 1024]

    def num_neighboring_slices(self):
        return self.IMG_SIZE[2]

    def summary(self, additional_vars=None):
        summary_attrs = ["SLICE_SUBSET"]
        super().summary(summary_attrs)


class DeeplabV3_2_5DConfig(DeeplabV3Config):
    """2.5D DeeplabV3+."""

    IMG_SIZE = (288, 288, 3)

    def __init__(self, state="training", create_dirs=True):
        warnings.warn("UNet2_5DConfig is deprecated. Use UNet2DConfig instead.", DeprecationWarning)
        super().__init__(state, create_dirs)

    def num_neighboring_slices(self):
        return self.IMG_SIZE[2]


class AnisotropicUNetConfig(Config):
    """2D Anisotropic U-Net."""

    MODEL_NAME = "anisotropic_unet"

    IMG_SIZE = (288, 72, 1)

    INITIAL_LEARNING_RATE = 2e-2
    DROP_FACTOR = 0.85
    DROP_RATE = 1.0
    TRAIN_BATCH_SIZE = 60

    DEPTH = 6
    NUM_FILTERS = None

    KERNEL_SIZE = (7, 3)

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)

    def summary(self, additional_vars=None):
        summary_attrs = ["DEPTH", "NUM_FILTERS", "KERNEL_SIZE"]
        super().summary(summary_attrs)


class RefineNetConfig(Config):
    """Configuration for RefineNet architecture as suggested by paper below
    http://openaccess.thecvf.com/content_cvpr_2017/papers/Lin_RefineNet_Multi
    -Path_Refinement_CVPR_2017_paper.pdf
    """

    MODEL_NAME = "refinenet"

    INITIAL_LEARNING_RATE = 1e-3

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)


class FCDenseNetConfig(Config):
    MODEL_NAME = "FCDenseNet"

    DEPTH = 5
    NUM_LAYERS = (5,)
    NUM_FILTERS = (12,)
    NUM_FILTERS_HEAD_CONV = 48
    DROPOUT = 0.2

    def __init__(self, state="training", create_dirs=True):
        super().__init__(self.MODEL_NAME, state, create_dirs=create_dirs)


def _check_and_coerce_cfg_value_type(replacement, original, full_key):
    """Checks that `replacement`, which is intended to replace `original` is of
    the right type. The type is correct if it matches exactly or is one of a few
    cases in which the type can be easily coerced.
    """
    original_type = type(original)
    replacement_type = type(replacement)

    # TODO: Convert all to have non-None values by default.
    if isinstance(original, type(None)):
        return replacement

    # The types must match (with some exceptions)
    if replacement_type == original_type:
        return replacement

    # Cast replacement from from_type to to_type if the replacement and original
    # types match from_type and to_type
    def conditional_cast(from_type, to_type):
        if replacement_type == from_type and original_type == to_type:
            return True, to_type(replacement)
        else:
            return False, None

    # Conditionally casts
    # list <-> tuple
    casts = [(tuple, list), (list, tuple)]

    for (from_type, to_type) in casts:
        converted, converted_value = conditional_cast(from_type, to_type)
        if converted:
            return converted_value

    raise ValueError(
        "Type mismatch ({} vs. {}) with values ({} vs. {}) for config "
        "key: {}".format(original_type, replacement_type, original, replacement, full_key)
    )


def _assert_with_logging(cond, msg):
    if not cond:
        logger.debug(msg)
    assert cond, msg


def _error_with_logging(cond, msg, error_type=ValueError):
    if not cond:
        logger.error(msg)
        raise error_type(msg)


SUPPORTED_CONFIGS = [
    UNetConfig,
    SegnetConfig,
    DeeplabV3Config,
    ResidualUNet,
    AnisotropicUNetConfig,
    RefineNetConfig,
    UNet3DConfig,
    UNet2_5DConfig,
    DeeplabV3_2_5DConfig,
    FCDenseNetConfig,
    ContextEncoderConfig,
    ContextUNetConfig,
    ContextInpaintingConfig,
    ContextSegmentationConfig,
]


def get_config(config_cp_save_tag: str, create_dirs: bool = True):
    """Get config using config_cp_save_tag

    Args:
        config_cp_save_tag: config cp_save_tag
        create_dirs: if directory should be created

    Return:
        Config: A config.
    """
    from medsegpy.modeling.meta_arch.build import LEGACY_MODEL_NAMES

    if config_cp_save_tag in LEGACY_MODEL_NAMES:
        config_cp_save_tag = LEGACY_MODEL_NAMES[config_cp_save_tag]

    configs = SUPPORTED_CONFIGS
    for config in configs:
        if config.MODEL_NAME == config_cp_save_tag:
            c = config(create_dirs=create_dirs)
            return c

    raise ValueError("config %s not found" % config_cp_save_tag)


def get_model_name(cfg_filename: str):
    """Get "MODEL_NAME" attribute from config file.

    Args:
        cfg_filename: filepath to INI or YAML file where config is stored

    Returns:
        str: The model name.
    """
    vars_dict = Config.parse_base_config(cfg_filename)
    return vars_dict["MODEL_NAME"] if "MODEL_NAME" in vars_dict else vars_dict["CP_SAVE_TAG"]



"""Convolutional Layers

The following layers are possible replacements for common convolutional
layers.
"""
import keras.backend as K
import tensorflow as tf
from keras.layers.convolutional import Conv2D


class ConvStandardized2D(Conv2D):
    """
    Implements a 2D convolutional layer with weight standardization.

    Reference:
        Qiao et al. "Micro-batch training with batch-channel normalization
        and weight standardization". arXiv preprint arXiv:1903.10520. 2019.
    """

    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding="valid",
        data_format=None,
        dilation_rate=(1, 1),
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super(ConvStandardized2D, self).__init__(
            filters,
            kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )

    def call(self, inputs):
        """
        This call function is essentially a copy of the call function
        for the private class _Conv, taking only the parts relevant
        for 2D convolution, and adding weight standardization at the
        very beginning of the function.
        """
        # Weight standardization
        # -- copied from "https://github.com/joe-siyuan-qiao/WeightStandardization"
        kernel = self.kernel
        kernel_mean = tf.math.reduce_mean(kernel, axis=[0, 1, 2], keepdims=True, name="kernel_mean")
        kernel = kernel - kernel_mean
        kernel_std = K.std(kernel, axis=[0, 1, 2], keepdims=True)
        kernel = kernel / (kernel_std + 1e-5)

        # Copied from call() in _Conv class
        outputs = K.conv2d(
            inputs,
            kernel,
            strides=self.strides,
            padding=self.padding,
            data_format=self.data_format,
            dilation_rate=self.dilation_rate,
        )

        if self.use_bias:
            outputs = K.bias_add(outputs, self.bias, data_format=self.data_format)

        if self.activation is not None:
            return self.activation(outputs)
        return outputs



import json
import logging
import multiprocessing as mp
import os
import random
import time
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Sequence, Tuple, Union

import h5py
import numpy as np

from medsegpy.config import Config

from .catalog import MetadataCatalog
from .fname_parsers import OAISliceWise

logger = logging.getLogger(__name__)

__all__ = [
    "get_generator",
    "GeneratorState",
    "Generator",
    "OAIGenerator",
    "OAI3DGenerator",
    "OAI3DBlockGenerator",
    "OAI3DGeneratorFullVolume",
    "CTGenerator",
]


class GeneratorState(Enum):
    """Defines the generator state for training/testing networks"""

    TRAINING = 1
    VALIDATION = 2
    TESTING = 3


def get_generator(cfg: Config):
    """Get generator based on config `"TAG"` value."""
    for generator in [
        # OAI supported generators.
        OAIGenerator,
        OAI3DGenerator,
        OAI3DBlockGenerator,
        OAI3DGeneratorFullVolume,
        # abCT supported generators.
        CTGenerator,
    ]:
        try:
            gen = generator(cfg)
            if gen:
                return gen
        except ValueError:
            continue

    raise ValueError("No generator found for tag `%s`" % cfg.TAG)


class Generator(ABC):
    """Abstract class for data generator for network training and testing."""

    SUPPORTED_TAGS = [""]

    def __init__(self, cfg: Config):
        warnings.warn("Generator is deprecated, use DataLoader instead", DeprecationWarning)

        if cfg.TAG not in self.SUPPORTED_TAGS:
            raise ValueError("Tag mismatch: config must have tag in %s" % self.SUPPORTED_TAGS)
        self.config = cfg
        self.fname_parser = OAISliceWise()

    @abstractmethod
    def _load_inputs(self, data_path, file) -> Tuple[np.ndarray, np.ndarray]:
        """Load inputs and ground truth from `os.path.join(data_path, file)`.

        Args:
            data_path (str): Root directory
            file (str): A filename

        Returns:
            Tuple[ndarray, ndarray]: inputs, outputs
        """
        pass

    @abstractmethod
    def img_generator(self, state: GeneratorState):
        """Data generator that yields `(inputs, outputs)`.

        Use with `model.fit_generator()`.

        Args:
            state (GeneratorState): `GeneratorState.TRAINING` or
                `GeneratorState.VALIDATION`

        Yields:
            Tuple[ndarray, ndarray]: batched inputs, batched outputs
        """
        accepted_states = [GeneratorState.TRAINING, GeneratorState.VALIDATION]
        if state not in accepted_states:
            raise ValueError("Generator must be either {}".format(accepted_states))

    @abstractmethod
    def img_generator_test(self, model=None):
        """Data generator that yields data during inference.

        Args:
            model (keras.models.Model): A Keras model to use for inference.

        Yields:
            inputs, outputs (ground truth), predictions, scan id, time elapsed.
                predictions and time elapsed are `None` if `model` not
                specified. Volumes will have shape `DxHxW`.
        """
        pass

    @abstractmethod
    def summary(self):
        """Log summary based on `self.cfg.STATE`."""
        pass

    @abstractmethod
    def num_steps(self) -> Tuple[int, int]:
        """Returns number of steps per epoch for training and validation.

        By default it calculates as `len(elements) // batch_size`. Elements are
        the unique elements that are batched together after structuring.

        Use for args `{train,validation}_steps` args in Keras `model.fit()`.

        Returns:
            Tuple[int, int]: training steps, validation steps
        """
        return 0, 0

    @abstractmethod
    def num_examples(self, state: GeneratorState) -> int:
        """Number of examples in this training set."""
        pass

    @abstractmethod
    def num_scans(self, state: GeneratorState) -> int:
        """Number of scans corresponding to given state."""
        pass

    def sort_files(self, files) -> List[str]:
        """Sort files by name.

        Args:
            files (Sequence[str]): Files to load.

        Returns:
            List[str]: Sorted files.
        """

        def argsort(seq):
            return sorted(range(len(seq)), key=seq.__getitem__)

        file_id = [None] * len(files)
        for cnt1 in range(len(files)):
            file_id[cnt1] = self.fname_parser.get_file_id(files[cnt1])

        order = argsort(file_id)

        return [files[cnt1] for cnt1 in order]

    def _add_file(
        self,
        file: str,
        unique_filenames: Sequence[str],
        pids: Sequence[Union[str, int]],
        augment_data: bool,
    ):
        """Check if file should be added to list of files.

        This is useful when wanting to filter out augmented data, specific
        subject/patient ids, etc.

        Args:
            file (str): Filename or filepath.
            unique_filenames (Sequence[str]): Running list of unique filenames
                to add.
            pids (Sequence[str/int]): Patient ids to include. If `None`, check
                is not made.
            augment_data (bool): If `True`, include augmented data.

        Returns:
            bool: If `True`, file should be added.
        """
        should_add_file = file not in unique_filenames
        file_info = self.fname_parser.get_file_info(file)

        if pids is not None:
            contains_pid = [str(x) in file for x in pids]

            # if any pid is included, only 1 can be included
            assert sum(contains_pid) in {0, 1}
            contains_pid = any(contains_pid)

            should_add_file &= contains_pid

        if not augment_data:
            if "aug" not in file_info.keys():
                warnings.simplefilter("once")
                warnings.warn(
                    "Augmentation index not found in filename. "
                    "Will add all files regardless of name."
                )
            else:
                should_add_file &= file_info["aug"] == 0

        return should_add_file

    def _add_background_labels(self, segs: np.ndarray):
        """Add index for background labels based on segmentations.

        Args:
            segs (ndarray): A binary array of shape (...)xC- i.e. last dimension
                is class dimension.

        Returns:
            ndarray: Last dim has length `C+1`, where `x[...,0]` is background.
        :return:
        """
        all_tissues = np.sum(segs, axis=-1, dtype=np.bool)
        background = np.asarray(~all_tissues, dtype=np.float)
        background = background[..., np.newaxis]
        seg_total = np.concatenate([background, segs], axis=-1)

        return seg_total

    def _img_generator_base_info(self, state: GeneratorState):
        """Get base info for TRAINING, VALIDATION, or TESTING generator states.

        Args:
            state (GeneratorState):

        Returns:
            dict: Basic information pertaining to this GeneratorState.
                Includes the following keys:
                * "data_path_or_files"
                * "batch_size"
                * "shuffle_epoch"
                * "pids"
                * "augment_data"
        """
        config = self.config
        if state == GeneratorState.TRAINING:
            # training
            data_path_or_files = (
                config.__CV_TRAIN_FILES__
                if config.USE_CROSS_VALIDATION
                else MetadataCatalog.get(config.TRAIN_DATASET).scan_root
            )
            batch_size = config.TRAIN_BATCH_SIZE
            shuffle_epoch = True
            pids = config.PIDS
            augment_data = config.AUGMENT_DATA
        elif state == GeneratorState.VALIDATION:
            # validation
            data_path_or_files = (
                config.__CV_VALID_FILES__
                if config.USE_CROSS_VALIDATION
                else MetadataCatalog.get(config.VAL_DATASET).scan_root
            )
            batch_size = config.VALID_BATCH_SIZE
            shuffle_epoch = False
            pids = None
            augment_data = False
        elif state == GeneratorState.TESTING:
            data_path_or_files = (
                config.__CV_TEST_FILES__
                if config.USE_CROSS_VALIDATION
                else MetadataCatalog.get(config.TEST_DATASET).scan_root
            )
            batch_size = config.TEST_BATCH_SIZE
            shuffle_epoch = False
            pids = None
            augment_data = False

        base_info = {
            "data_path_or_files": data_path_or_files,
            "batch_size": batch_size,
            "shuffle_epoch": shuffle_epoch,
            "pids": pids,
            "augment_data": augment_data,
        }

        return base_info


class OAIGenerator(Generator):
    """Generator for data stored as 2D slices without structuring.

    In this generator, each slice is treated as a different element. For 2.5D
    networks, it loads sequential slices in the channel dimension.
    It should be used for 2D and 2.5D networks.

    Data for this generator must be stored in the medsegpy 2D dataset format.
    See https://ad12.github.io/MedSegPy/_build/html/tutorials/datasets.html#data-format  # noqa
    for more information.
    """

    SUPPORTED_TAGS = ["oai_aug", "oai", "oai_2d", "oai_aug_2d", "oai_2.5d", "oai_aug_2.5d"]
    __EXPECTED_IMG_SIZE_DIMS__ = 3

    def img_generator(self, state: GeneratorState):
        super().img_generator(state)

        config = self.config
        img_size = config.IMG_SIZE
        tissues = config.CATEGORIES
        include_background = config.INCLUDE_BACKGROUND
        num_neighboring_slices = config.num_neighboring_slices()

        base_info = self._img_generator_base_info(state)
        batch_size = base_info["batch_size"]
        shuffle_epoch = base_info["shuffle_epoch"]
        files, batches_per_epoch, max_slice_num = self._calc_generator_info(state)

        total_classes = config.get_num_classes()
        mask_size = img_size[:-1] + (total_classes,)
        x = np.zeros((batch_size,) + img_size)
        y = np.zeros((batch_size,) + mask_size)

        rand_obj = random.Random(config.SEED)
        if config.SEED is not None:
            files = self.sort_files(files)

        while True:
            if shuffle_epoch:
                rand_obj.shuffle(files)
            else:
                files = self.sort_files(files)

            for batch_cnt in range(batches_per_epoch):
                for file_cnt in range(batch_size):
                    file_ind = batch_cnt * batch_size + file_cnt
                    filepath = files[file_ind]

                    im, seg = self._load_input_helper(
                        filepath=filepath,
                        tissues=tissues,
                        num_neighboring_slices=num_neighboring_slices,
                        max_slice_num=max_slice_num,
                        include_background=include_background,
                    )

                    assert im.shape == img_size, "Input shape mismatch. Expected %s, got %s" % (
                        img_size,
                        im.shape,
                    )
                    assert seg.shape == mask_size, "Ouput shape mismatch. Expected %s, got %s" % (
                        mask_size,
                        seg.shape,
                    )

                    x[file_cnt, ...] = im
                    y[file_cnt, ...] = seg

                yield (x, y)

    def img_generator_test(self, model=None):
        config = self.config
        img_size = config.IMG_SIZE
        tissues = config.CATEGORIES
        include_background = config.INCLUDE_BACKGROUND
        num_neighboring_slices = config.num_neighboring_slices()

        base_info = self._img_generator_base_info(GeneratorState.TESTING)
        batch_size = base_info["batch_size"]

        if len(img_size) != self.__EXPECTED_IMG_SIZE_DIMS__:
            raise ValueError("Image size must be %dD" % self.__EXPECTED_IMG_SIZE_DIMS__)

        total_classes = config.get_num_classes()
        mask_size = img_size[:-1] + (total_classes,)

        files, batches_per_epoch, _ = self._calc_generator_info(GeneratorState.TESTING)
        files = self.sort_files(files)
        scan_id_to_files = self._map_files_to_scan_id(files)
        scan_ids = sorted(scan_id_to_files.keys())

        for scan_id in list(scan_ids):
            scan_id_files = scan_id_to_files[scan_id]
            num_slices = len(scan_id_files)

            x = np.zeros((num_slices,) + img_size)
            y = np.zeros((num_slices,) + mask_size)

            for fcount in range(num_slices):
                filepath = scan_id_files[fcount]

                # Make sure that this pid is actually in the filename
                assert scan_id in filepath, "scan_id missing: id %s not in %s" % (scan_id, filepath)

                im, seg_total = self._load_input_helper(
                    filepath=filepath,
                    tissues=tissues,
                    num_neighboring_slices=num_neighboring_slices,
                    max_slice_num=num_slices,
                    include_background=include_background,
                )

                x[fcount, ...] = im
                y[fcount, ...] = seg_total

            recon = None
            if model:
                start_time = time.time()
                recon = model.predict(x, batch_size=batch_size)
                time_elapsed = time.time() - start_time
                x, y, recon = self._reformat_testing_scans((x, y, recon))
            else:
                x, y = self._reformat_testing_scans((x, y))

            yield (x, y, recon, scan_id, time_elapsed)

    def _reformat_testing_scans(self, vols: Sequence[np.ndarray]):
        """Reformat each volume in `vols` into DxHxW shape.

        Args:
            vols (Sequence[ndarray]): Volumes to be reformatted.

        Returns:
            Tuple[ndarray]: Volumes reformatted in DxHxW shape.
        """
        return tuple(vols)

    def _map_files_to_scan_id(self, files):
        """Map file names to scan_ids (patient-id_timepoint-id).

        Args:
            files: An iterable of files

        Returns:
            dict: scan_id -> files
        """
        scan_id_files = dict()
        for f in files:
            filename = os.path.basename(f)
            file_info = self.fname_parser.get_file_info(filename)
            scan_id = file_info["scanid"]

            if scan_id in scan_id_files.keys():
                scan_id_files[scan_id].append(f)
            else:
                scan_id_files[scan_id] = [f]

        for k in scan_id_files.keys():
            scan_id_files[k] = sorted(scan_id_files[k])

        return scan_id_files

    def _load_input_helper(
        self, filepath, tissues, num_neighboring_slices, max_slice_num, include_background
    ):
        """

        :param filepath:
        :param tissues:
        :param num_neighboring_slices: The number of slices to load
            None for 2D networks
        :param max_slice_num:
        :param include_background:
        :return:
        """
        # check that fname contains a dirpath
        assert os.path.dirname(filepath) != ""

        if num_neighboring_slices:
            im, seg = self._load_neighboring_slices(
                num_slices=num_neighboring_slices, filepath=filepath, max_slice=max_slice_num
            )
        else:
            im, seg = self._load_inputs(os.path.dirname(filepath), os.path.basename(filepath))

        # support multi class
        if len(tissues) > 1:
            seg_tissues = self._compress_multi_class_mask(seg, tissues)
        else:
            seg_tissues = seg[..., 0, tissues]
            if isinstance(
                tissues[0], list
            ):  # if tissues are supposed to be summed (like tibial cartilage
                # and meniscus)
                seg_tissues = np.sum(seg_tissues, axis=-1)

        seg_total = seg_tissues

        # if considering background, add class
        # background should mark every other pixel that is not already
        # accounted for in segmentation
        if include_background:
            seg_total = self._add_background_labels(seg_tissues)

        return im, seg_total

    def _compress_multi_class_mask(self, seg, tissues):
        o_seg = []

        for t_inds in tissues:
            c_seg = seg[..., 0, t_inds]
            if c_seg.ndim == 3:
                c_seg = np.sum(c_seg, axis=-1)
            o_seg.append(c_seg)

        return np.stack(o_seg, axis=-1)

    def _load_neighboring_slices(self, num_slices, filepath, max_slice):
        """
        Assumes that there are at most slices go from 1-max_slice
        :param num_slices:
        :param filepath:
        :return:
        """
        data_path, filename = (os.path.dirname(filepath), os.path.basename(filepath))

        d_slice = num_slices // 2
        filename_split = filename.split("_")
        central_slice_no = int(filename_split[-1])

        slice_nos = np.asarray(
            list(range(central_slice_no - d_slice, central_slice_no + d_slice + 1))
        )
        slice_nos[slice_nos < 1] = 1
        slice_nos[slice_nos > max_slice] = max_slice

        assert len(slice_nos) == num_slices

        base_filename = "_".join(filename_split[:-1]) + "_%03d"

        ims = []
        segs = []
        for i in range(num_slices):
            slice_no = slice_nos[i]
            slice_filename = base_filename % slice_no

            im, seg = self._load_inputs(data_path, slice_filename)
            ims.append(im)
            segs.append(seg)

        # segmentation is central slice segmentation
        im = np.squeeze(np.stack(ims, axis=-1))
        # im = np.transpose(im, (1, 2, 0))
        seg = segs[len(segs) // 2]

        return im, seg

    def _load_inputs(self, data_path: str, file: str):
        im_path = "%s/%s.im" % (data_path, file)
        with h5py.File(im_path, "r") as f:
            im = f["data"][:]
            if len(im.shape) == 2:
                im = im[..., np.newaxis]

        seg_path = "%s/%s.seg" % (data_path, file)
        with h5py.File(seg_path, "r") as f:
            seg = f["data"][:].astype("float32")

        assert len(im.shape) == 3
        assert len(seg.shape) == 4 and seg.shape[-2] == 1

        return im, seg

    def _parse_coco_json(self, coco_file: str):
        with open(coco_file) as f:
            data = json.load(f)
        files = [image_data["file_name"] for image_data in data["images"]]

        return files

    def _parse_txt_file(self, txt_file: str):
        with open(txt_file) as f:
            files = f.readlines()
        files = [f.strip() for f in files]
        return files

    def _parse_filepaths(self, data_path_or_files):
        if type(data_path_or_files) is str:
            if os.path.isdir(data_path_or_files):
                data_path = data_path_or_files
                files = os.listdir(data_path)
                filepaths = [os.path.join(data_path, f) for f in files]
            elif data_path_or_files.endswith(".json"):
                filepaths = self._parse_coco_json(data_path_or_files)
            elif data_path_or_files.endswith(".txt"):
                filepaths = self._parse_txt_file(data_path_or_files)
            else:
                raise ValueError("Unknown file to parse {}".format(data_path_or_files))
        elif type(data_path_or_files) is list:
            filepaths = data_path_or_files
        else:
            raise ValueError("data_path_or_files must be type str or list")

        return filepaths

    def _calc_generator_info(self, state: GeneratorState):
        base_info = self._img_generator_base_info(state)
        data_path_or_files = base_info["data_path_or_files"]
        batch_size = base_info["batch_size"]
        pids = base_info["pids"]
        augment_data = base_info["augment_data"]

        filepaths = self._parse_filepaths(data_path_or_files)

        unique_filepaths = {}  # use dict to avoid having to reconstruct set every time

        # track the largest slice number that we see
        # assume it is the same for all scans
        max_slice_num = 0

        for fp in filepaths:
            fp, _ = os.path.splitext(fp)
            _, filename = os.path.dirname(fp), os.path.basename(fp)

            if self._add_file(fp, unique_filepaths, pids, augment_data):
                file_info = self.fname_parser.get_file_info(filename)
                if max_slice_num < file_info["slice"]:
                    max_slice_num = file_info["slice"]

                unique_filepaths[fp] = fp

        files = list(unique_filepaths.keys())

        # Set total number of files based on argument for limiting training size
        nfiles = len(files)

        batches_per_epoch = nfiles // batch_size

        return files, batches_per_epoch, max_slice_num

    def summary(self):
        config = self.config

        if config.STATE == "training":
            train_files, train_batches_per_epoch, _ = self._calc_generator_info(
                GeneratorState.TRAINING
            )
            valid_files, valid_batches_per_epoch, _ = self._calc_generator_info(
                GeneratorState.VALIDATION
            )

            # Get number of subjects training and validation sets
            train_pids = []
            for f in train_files:
                file_info = self.fname_parser.get_file_info(os.path.basename(f))
                train_pids.append(file_info["pid"])

            valid_pids = []
            for f in valid_files:
                file_info = self.fname_parser.get_file_info(os.path.basename(f))
                valid_pids.append(file_info["pid"])

            num_train_subjects = len(set(train_pids))
            num_valid_subjects = len(set(valid_pids))

            logger.info(
                "INFO: Train size: %d slices (%d subjects), batch size: %d"
                % (len(train_files), num_train_subjects, self.config.TRAIN_BATCH_SIZE)
            )
            logger.info(
                "INFO: Valid size: %d slices (%d subjects), batch size: %d"
                % (len(valid_files), num_valid_subjects, self.config.VALID_BATCH_SIZE)
            )
            logger.info("INFO: Image size: %s" % (self.config.IMG_SIZE,))
            logger.info("INFO: Image types included in training: %s" % (self.config.FILE_TYPES,))
        else:  # config in Testing state
            test_files, test_batches_per_epoch, _ = self._calc_generator_info(
                GeneratorState.TESTING
            )
            scanset_info = self.__get_scanset_data__(test_files)

            logger.info(
                "INFO: Test size: %d slices, batch size: %d, "
                "# subjects: %d, # scans: %d"
                % (
                    len(test_files),
                    config.TEST_BATCH_SIZE,
                    len(scanset_info["pid"]),
                    len(scanset_info["scanid"]),
                )
            )
            if not config.USE_CROSS_VALIDATION:
                logger.info("Test path: %s" % config.TEST_PATH)

    def num_scans(self, state):
        files, _, _ = self._calc_generator_info(state)
        scanset_info = self.__get_scanset_data__(files)
        return len(scanset_info["scanid"])

    def __get_scanset_data__(self, files, keys=("pid", "scanid")):
        info_dict = dict()
        for k in keys:
            info_dict[k] = []

        for f in files:
            file_info = self.fname_parser.get_file_info(os.path.basename(f))
            for k in keys:
                info_dict[k].append(file_info[k])

        for k in keys:
            info_dict[k] = list(set(info_dict[k]))

        return info_dict

    def num_steps(self):
        _, train_batches_per_epoch, _ = self._calc_generator_info(GeneratorState.TRAINING)
        _, valid_batches_per_epoch, _ = self._calc_generator_info(GeneratorState.VALIDATION)

        return train_batches_per_epoch, valid_batches_per_epoch

    def num_examples(self, state: GeneratorState):
        files, _, _ = self._calc_generator_info(state)
        return len(files)


class OAI3DGenerator(OAIGenerator):
    """Generator for training 3D networks where data is stored as 2D slices."""

    SUPPORTED_TAGS = ["oai_3d"]
    __EXPECTED_IMG_SIZE_DIMS__ = 4

    def __init__(self, config: Config):
        if config.TAG == "oai_3d_block-train_full-test":
            assert config.testing, "Config must be in testing state to use tag %s" % config.TAG

        super().__init__(config)

    def __validate_img_size__(self, slices_per_scan):
        # only accept image sizes where slices can be perfectly disjoint
        if len(self.config.IMG_SIZE) != self.__EXPECTED_IMG_SIZE_DIMS__:
            raise ValueError(
                "`IMG_SIZE` must be in format (y, x, z, #channels) - got %s"
                % str(self.config.IMG_SIZE)
            )

        input_volume_num_slices = self.config.IMG_SIZE[2]
        if input_volume_num_slices == 1:
            raise ValueError("For 2D/2.5D networks, use `OAIGenerator`")

        if slices_per_scan % input_volume_num_slices != 0:
            raise ValueError(
                "All input volumes must be disjoint. "
                "%d slices per scan, but %d slices in input"
                % (slices_per_scan, input_volume_num_slices)
            )

    def __get_corresponding_files__(self, fname: str):
        num_slices = self.config.IMG_SIZE[2]
        file_info = self.fname_parser.get_file_info(fname)
        f_slice = file_info["slice"]

        volume_index = int((f_slice - 1) / num_slices)

        base_info = {
            "pid": file_info["pid"],
            "timepoint": file_info["timepoint"],
            "aug": file_info["aug"],
            "slice": file_info["slice"],
        }

        # slices are 1-indexed
        slices = list(range(volume_index * num_slices + 1, (volume_index + 1) * num_slices + 1))
        slice_fnames = []
        for s in slices:
            base_info["slice"] = s
            slice_fnames.append(self.fname_parser.get_fname(base_info))

        return slice_fnames

    def _load_input_helper(
        self, filepath, tissues, num_neighboring_slices, max_slice_num, include_background
    ):
        # check that fname contains a dirpath
        assert os.path.dirname(filepath) != ""

        if num_neighboring_slices:
            im, seg = self.__load_corresponding_slices__(filepath)
        else:
            raise ValueError("`num_neighboring_slices` must be initialized")

        # support multi class
        if len(tissues) > 1:
            seg_tissues = self._compress_multi_class_mask(seg, tissues)
        else:
            seg_tissues = seg[..., 0, tissues]
            if isinstance(tissues[0], list):  # if tissues are supposed to be summed
                # (like tibial cartilage and meniscus)
                seg_tissues = np.sum(seg_tissues, axis=-1)
        seg_total = seg_tissues

        # if considering background, add class
        # background should mark every other pixel that is not already
        # accounted for in segmentation
        if include_background:
            seg_total = self._add_background_labels(seg_tissues)

        return im, seg_total

    def _compress_multi_class_mask(self, seg, tissues):
        o_seg = []

        for t_inds in tissues:
            c_seg = seg[..., t_inds, 0]
            if c_seg.ndim == 4:
                c_seg = np.sum(c_seg, axis=-1)
            o_seg.append(c_seg)

        return np.stack(o_seg, axis=-1)

    def __load_corresponding_slices__(self, filepath):
        """Loads volume that slice (filepath) corresponds to.

        e.g. If each input volume consists of 4 slices, slices 1-4 will be
        in the same volume. If `filepath` corresponds to slice 2, the volume
        (slices 1-4) and corresponding segmentations will be returned.

        :param filepath: Filepath corresponding to single slice
        :return: tuple of 2 3d numpy array corresponding to 3d input volume and
            3d segmentation volume (im_vol, seg_vol)
        """
        data_path, filename = (os.path.dirname(filepath), os.path.basename(filepath))
        corresponding_files = self.__get_corresponding_files__(filename)

        ims = []
        segs = []
        for f in corresponding_files:
            im, seg = self._load_inputs(data_path, f)
            ims.append(im)
            segs.append(seg)

        im_vol = np.stack(ims, axis=-1)
        seg_vol = np.stack(segs, axis=-1)
        im_vol = np.squeeze(im_vol)
        im_vol = im_vol[..., np.newaxis]
        seg_vol = np.transpose(np.squeeze(seg_vol), [0, 1, 3, 2])
        seg_vol = seg_vol[..., np.newaxis]

        assert im_vol.shape == self.config.IMG_SIZE, "Loaded volume of size %s. Expected %s" % (
            im_vol.shape,
            self.config.IMG_SIZE,
        )

        return im_vol, seg_vol

    def _calc_generator_info(self, state: GeneratorState):
        base_info = self._img_generator_base_info(state)
        data_path_or_files = base_info["data_path_or_files"]
        batch_size = base_info["batch_size"]
        pids = base_info["pids"]
        augment_data = base_info["augment_data"]

        filepaths = self._parse_filepaths(data_path_or_files)

        unique_filepaths = {}  # use dict to avoid having to reconstruct set every time

        # track the largest slice number that we see
        # assume it is the same for all scans
        slice_ids = []

        for fp in filepaths:
            fp, _ = os.path.splitext(fp)
            _, filename = os.path.dirname(fp), os.path.basename(fp)

            if self._add_file(fp, unique_filepaths, pids, augment_data):
                file_info = self.fname_parser.get_file_info(filename)
                slice_ids.append(file_info["slice"])

                unique_filepaths[fp] = fp

        files = list(unique_filepaths.keys())

        min_slice_id = min(slice_ids)
        max_slice_id = max(slice_ids)

        # validate image size
        self.__validate_img_size__(max_slice_id - min_slice_id + 1)

        # Remove files corresponding to the same volume
        # e.g. If input volume has 4 slices, slice 1 and 2 will be part
        # of the same volume.
        # We only include file corresponding to slice 1, because other files
        # are accounted for by default
        slices_to_include = range(min_slice_id, max_slice_id + 1, self.config.IMG_SIZE[2])
        files_refined = []
        for filepath in files:
            fname = os.path.basename(filepath)
            f_info = self.fname_parser.get_file_info(fname)
            if f_info["slice"] in slices_to_include:
                files_refined.append(filepath)

        files = list(set(files_refined))

        # Set total number of volumes based on argument for limiting
        # training size
        nvolumes = len(files)

        batches_per_epoch = nvolumes // batch_size

        return files, batches_per_epoch, max_slice_id

    def _add_file(self, file: str, unique_filenames, pids, augment_data: bool):
        add_file = super()._add_file(file, unique_filenames, pids, augment_data)

        # If only subset of slices should be selected, then return
        if hasattr(self.config, "SLICE_SUBSET") and self.config.SLICE_SUBSET is not None:
            file_info = self.fname_parser.get_file_info(file)
            add_file &= file_info["slice"] in range(
                self.config.SLICE_SUBSET[0], self.config.SLICE_SUBSET[1] + 1
            )

        return add_file


class OAI3DGeneratorFullVolume(OAI3DGenerator):
    SUPPORTED_TAGS = ["oai_3d_block-train_full-vol-test", "full-vol-test", "oai_3d_full-vol-test"]

    def __init__(self, config: Config):
        if not config.testing:
            raise ValueError("%s only available for testing 3D volumes" % self.__class__.__name__)

        super().__init__(config)

        # update image size for config to be total volume
        img_size = self.__get_img_dims__(GeneratorState.TESTING)
        assert len(img_size) == 3, "Loaded image size must be a tuple of 3 integers (y, x, #slices)"
        self.config.IMG_SIZE = img_size + self.config.IMG_SIZE[3:]

    def _img_generator_base_info(self, state: GeneratorState):
        assert state == GeneratorState.TESTING, "Only testing state is supported for this generator"
        return super()._img_generator_base_info(state)

    def __get_img_dims__(self, state: GeneratorState):
        base_info = self._img_generator_base_info(state)
        data_path_or_files = base_info["data_path_or_files"]
        pids = base_info["pids"]
        augment_data = base_info["augment_data"]

        if type(data_path_or_files) is str:
            data_path = data_path_or_files
            files = os.listdir(data_path)
            filepaths = [os.path.join(data_path, f) for f in files]
        elif type(data_path_or_files) is list:
            filepaths = data_path_or_files
        else:
            raise ValueError("data_path_or_files must be type str or list")

        unique_filepaths = {}  # use dict to avoid having to reconstruct set every time

        # track the largest slice number that we see
        # assume it is the same for all scans
        slice_ids = []

        for fp in filepaths:
            fp, _ = os.path.splitext(fp)
            _, filename = os.path.dirname(fp), os.path.basename(fp)

            if self._add_file(fp, unique_filepaths, pids, augment_data):
                file_info = self.fname_parser.get_file_info(filename)
                slice_ids.append(file_info["slice"])

                unique_filepaths[fp] = fp

        min_slice_id = min(slice_ids)
        max_slice_id = max(slice_ids)
        num_slices = max_slice_id - min_slice_id + 1

        # load 1 file to get inplane resolution
        filepath = list(unique_filepaths.keys())[0]
        im_vol, _ = self._load_inputs(os.path.dirname(filepath), os.path.basename(filepath))
        return im_vol.shape[:2] + (num_slices,)

    def _reformat_testing_scans(self, vols):
        vols_updated = []
        for v in vols:
            assert (
                v.ndim == 5 and v.shape[0] == 1 and v.shape[1:] == self.config.IMG_SIZE
            ), "img dims must be %s" % str((1,) + self.config.IMG_SIZE)
            v_updated = np.squeeze(v, axis=0)
            v_updated = np.transpose(v_updated, [2, 0, 1, 3])
            vols_updated.append(v_updated)
        return tuple(vols_updated)

    def num_steps(self):
        raise ValueError("This method is not supported for a testing-only generator")


class OAI3DBlockGenerator(OAI3DGenerator):
    """Generator for structuring 2D data into 3D blocks."""

    SUPPORTED_TAGS = ["oai_3d_block"]

    def __init__(self, config):
        super().__init__(config=config)
        self._cached_data = dict()

    def cached_data(self, state: GeneratorState):
        if state not in self._cached_data.keys():
            start_time = time.time()
            logger.info("Computing %s blocks" % state.name)
            self._cached_data[state] = self._calc_generator_info(state)
            logger.info("%0.2f seconds" % (time.time() - start_time))

        return self._cached_data[state]

    def process_filepath(self, fp):
        dirpath = os.path.dirname(fp)
        fname = os.path.basename(fp)

        im, seg = self._load_inputs(data_path=dirpath, file=fname)
        im = np.squeeze(im)
        seg = np.squeeze(seg)
        assert (
            im.ndim == 2 and seg.ndim == 3
        ), "image must be 2D (Y,X) and segmentation must be 3D (Y,X,#masks)"

        info = self.fname_parser.get_file_info(fname)
        volume_id = info["volume_id"]
        slice_num = info["slice"]
        return volume_id, slice_num, im, seg

    def __load_all_volumes__(self, filepaths):
        parallelize = False
        if parallelize:
            with mp.Pool() as pool:
                loaded_data_list = pool.map(self.process_filepath, filepaths)
        else:
            loaded_data_list = []
            for fp in filepaths:
                loaded_data_list.append(self.process_filepath(fp))
        # sort list first by volume_id, then by slice_num
        loaded_data_list = sorted(loaded_data_list, key=(lambda x: (x[0], x[1])))

        scans_data = dict()
        for volume_id, _, im, seg in loaded_data_list:
            if volume_id not in scans_data.keys():
                scans_data[volume_id] = {"ims": [], "segs": []}

            scans_data[volume_id]["ims"].append(im)
            scans_data[volume_id]["segs"].append(seg)

        packaged_scan_volumes = {}
        for scan_id in scans_data.keys():
            im_vol = np.stack(scans_data[scan_id]["ims"], axis=-1)
            seg_vol = np.stack(scans_data[scan_id]["segs"], axis=-1)
            seg_vol = np.transpose(seg_vol, [0, 1, 3, 2])

            packaged_scan_volumes[scan_id] = (im_vol, seg_vol)

        return packaged_scan_volumes

    def unify_blocks(self, blocks, reshape_dims):
        # blocks are stored in decreasing order of fastest changing axes: x,y,z
        # inverse of blockify volume
        """
        z= 0:
        -----------------
        | 1 | 2 | 3 | 4 |
        -----------------
        | 5 | 6 | 7 | 8 |
        -----------------
        | 9 | 10| 11| 12|
        -----------------
        | 13| 14| 15| 16|
        -----------------
        z= 1:
        -----------------
        | 17| 18| 19| 20|
        -----------------
        | 21| 22| 23| 24|
        -----------------
        | 25| 26| 27| 28|
        -----------------
        | 29| 30| 31| 32|
        -----------------
        :param blocks: a list of tuples of numpy arrays
            [(im_block1, seg_block1), (im_block2, seg_block2)]
        :param reshape_dims:
        :return:
        """
        yb, xb, zb, _ = self.config.IMG_SIZE

        num_masks = blocks[0][1].shape[-1]
        im_volume = np.zeros(reshape_dims)
        seg_volume = np.zeros(reshape_dims + (num_masks,))

        ind = 0
        for z in range(0, reshape_dims[2], zb):
            for y in range(0, reshape_dims[1], yb):
                for x in range(0, reshape_dims[0], xb):
                    im, seg = blocks[ind]
                    im_volume[y : y + yb, x : x + xb, z : z + zb] = im
                    seg_volume[y : y + yb, x : x + xb, z : z + zb, :] = seg
                    ind += 1
        assert ind == len(blocks), "Blocks not appropriately portioned"

        return im_volume, seg_volume

    def blockify_volume(self, im_volume: np.ndarray, seg_volume: np.ndarray):
        assert im_volume.ndim == 3, "Dimension mismatch. im_volume must be 3D (y, x, z)."
        assert (
            seg_volume.ndim == 4
        ), "Dimension mismatch. seg_volume must be 4D (y, x, z, #classes)."

        yb, xb, zb, _ = self.config.IMG_SIZE
        expected_block_size = (yb, xb, zb)

        # verify that the sizes of im_volume and seg_volume are as expected
        assert (
            im_volume.shape == seg_volume.shape[:-1]
        ), "Input volume of size %s. Masks of size %s" % (im_volume.shape, seg_volume.shape)
        self.__validate_img_size__(im_volume.shape)

        expected_num_blocks = self.__calc_num_blocks__(im_volume.shape)
        blocks = []
        for z in range(0, im_volume.shape[2], zb):
            for y in range(0, im_volume.shape[1], yb):
                for x in range(0, im_volume.shape[0], xb):
                    im = im_volume[y : y + yb, x : x + xb, z : z + zb]
                    seg = seg_volume[y : y + yb, x : x + xb, z : z + zb, :]
                    assert (
                        im.shape[:3] == seg.shape[:3]
                    ), "Block shape mismatch. im_block %s, seg_block %s" % (im.shape, seg.shape)
                    assert (
                        im.shape[:3] == expected_block_size
                    ), "Block shape error. Expected %s, but got %s" % (
                        expected_block_size,
                        im.shape,
                    )
                    blocks.append((im, seg))
        assert len(blocks) == expected_num_blocks, "Expected %d blocks, got %d" % (
            expected_num_blocks,
            len(blocks),
        )

        return blocks

    def __blockify_volumes__(self, scan_volumes: dict):
        """Split all volumes into blocks of size `config.IMG_SIZE`.

        Throws error if volume is not perfectly divisible into blocks of size
        config.IMG_SIZE (Yi, Xi, Zi).

        :param scan_volumes: A dict of scan_ids --> (im_volume, seg_volume)
                                    im_volume: 3D numpy array (Y, X, Z)
                                    seg_volume: 4D numpy array (Y, X, Z, #masks)
        :return: a list of tuples im_volume and corresponding seg_volume blocks
                    im_volume block: a 3D numpy array of size Y/Yi, X/Xi, Z/Zi
                    seg_volume block: a 3D numpy array of size Y/Yi, X/Xi, Z/Zi, #masks  # noqa
        """
        ordered_keys = sorted(scan_volumes.keys())
        scan_to_blocks = dict()
        for scan_id in ordered_keys:
            im_volume, seg_volume = scan_volumes[scan_id]
            scan_to_blocks[scan_id] = self.blockify_volume(im_volume, seg_volume)

        return scan_to_blocks

    def __get_num_blocks__(self, scan_to_blocks):
        temp = []
        for k in scan_to_blocks:
            temp.extend(scan_to_blocks[k])
        return len(temp)

    def __calc_num_blocks__(self, total_volume_shape):
        num_blocks = 1
        for dim in range(3):
            assert total_volume_shape[dim] % self.config.IMG_SIZE[dim] == 0, (
                "Verify volumes prior to calling this function " "using `__validate_img_size__`"
            )
            num_blocks *= total_volume_shape[dim] // self.config.IMG_SIZE[dim]
        assert int(num_blocks) == num_blocks, "Num_blocks is %0.2f, must be an integer" % num_blocks

        return int(num_blocks)

    def _calc_generator_info(self, state: GeneratorState) -> Tuple[dict, int, dict]:
        base_info = self._img_generator_base_info(state)
        data_path_or_files = base_info["data_path_or_files"]
        batch_size = base_info["batch_size"]
        pids = base_info["pids"]
        augment_data = base_info["augment_data"]

        filepaths = self._parse_filepaths(data_path_or_files)

        unique_filepaths = {}  # use dict to avoid having to reconstruct set every time

        # track the largest slice number that we see - assume it is the same
        # for all scans
        slice_ids = []

        for fp in filepaths:
            fp, _ = os.path.splitext(fp)
            _, filename = os.path.dirname(fp), os.path.basename(fp)

            if self._add_file(fp, unique_filepaths, pids, augment_data):
                file_info = self.fname_parser.get_file_info(filename)
                slice_ids.append(file_info["slice"])

                unique_filepaths[fp] = fp

        files = list(unique_filepaths.keys())
        scan_to_volumes = self.__load_all_volumes__(files)

        # store original volume sizes in dict
        scan_to_im_size = dict()
        for scan in scan_to_volumes.keys():
            scan_to_im_size[scan] = scan_to_volumes[scan][0].shape

        scan_to_blocks = self.__blockify_volumes__(scan_to_volumes)

        # Set total number of volumes based on argument for limiting training
        # size
        nblocks = 0
        for k in scan_to_blocks.keys():
            nblocks += len(scan_to_blocks[k])

        batches_per_epoch = nblocks // batch_size

        return scan_to_blocks, batches_per_epoch, scan_to_im_size

    def __validate_img_size__(self, total_volume_shape):
        """
        Validate that image input size goes perfectly into volume
        :param total_volume_shape:
        :return:
        """
        super().__validate_img_size__(total_volume_shape[2])

        # check that all blocks will be disjoint
        # this means shape of total volume must be perfectly divisible into
        # cubes of size IMG_SIZE
        for dim in range(3):
            assert (
                total_volume_shape[dim] % self.config.IMG_SIZE[dim] == 0
            ), "Cannot divide volume of size %s to blocks of size %s" % (
                total_volume_shape,
                self.config.IMG_SIZE,
            )

    def img_generator_test(self, model=None):
        state = GeneratorState.TESTING
        base_info = self._img_generator_base_info(state)
        batch_size = base_info["batch_size"]

        config = self.config
        img_size = config.IMG_SIZE
        tissues = config.CATEGORIES
        include_background = config.INCLUDE_BACKGROUND

        scan_to_blocks, batches_per_epoch, scan_to_im_size = self.cached_data(state)

        total_classes = config.get_num_classes()
        mask_size = img_size[:-1] + (total_classes,)

        volume_ids = sorted(scan_to_blocks.keys())

        for vol_id in volume_ids:
            blocks = scan_to_blocks[vol_id]
            num_blocks = len(blocks)

            x = np.zeros((num_blocks,) + img_size)
            y = np.zeros((num_blocks,) + mask_size)

            for block_cnt in range(num_blocks):
                im, seg = blocks[block_cnt]
                if im.ndim == 3:
                    im = im[..., np.newaxis]

                seg = self.__format_seg_helper__(seg, tissues, include_background)
                assert im.shape == img_size, "Input shape mismatch. Expected %s, got %s" % (
                    img_size,
                    im.shape,
                )
                assert seg.shape == mask_size, "Ouput shape mismatch. Expected %s, got %s" % (
                    mask_size,
                    seg.shape,
                )

                x[block_cnt, ...] = im
                y[block_cnt, ...] = seg

            # reshape into original volume shape
            recon_vol = None
            ytrue_blocks = [(np.squeeze(x[b, ...]), y[b, ...]) for b in range(num_blocks)]
            im_vol, ytrue_vol = self.unify_blocks(ytrue_blocks, scan_to_im_size[vol_id])

            # reshape to expected output shape
            im_vol = np.transpose(im_vol, [2, 0, 1])
            im_vol = im_vol[..., np.newaxis]
            ytrue_vol = np.transpose(ytrue_vol, [2, 0, 1, 3])

            if model:
                start_time = time.time()
                recon = model.predict(x, batch_size=batch_size)
                time_elapsed = time.time() - start_time
                ypred_blocks = [(np.squeeze(x[b, ...]), recon[b, ...]) for b in range(num_blocks)]
                _, recon_vol = self.unify_blocks(ypred_blocks, scan_to_im_size[vol_id])
                recon_vol = np.transpose(recon_vol, [2, 0, 1, 3])

            yield (im_vol, ytrue_vol, recon_vol, vol_id, time_elapsed)

    def img_generator(self, state):
        accepted_states = [GeneratorState.TRAINING, GeneratorState.VALIDATION]
        if state not in accepted_states:
            raise ValueError("state must be either %s" % accepted_states)

        base_info = self._img_generator_base_info(state)
        batch_size = base_info["batch_size"]
        shuffle_epoch = base_info["shuffle_epoch"]

        config = self.config
        img_size = config.IMG_SIZE
        tissues = config.CATEGORIES
        include_background = config.INCLUDE_BACKGROUND

        scan_to_blocks, batches_per_epoch, _ = self.cached_data(state)

        total_classes = config.get_num_classes()
        mask_size = img_size[:-1] + (total_classes,)

        x = np.zeros((batch_size,) + img_size)
        y = np.zeros((batch_size,) + mask_size)

        blocks = []

        if config.SEED is not None:
            scan_ids = sorted(scan_to_blocks.keys())
        else:
            scan_ids = scan_to_blocks.keys()

        for k in scan_ids:
            blocks.extend(scan_to_blocks[k])

        rand_obj = random.Random(config.SEED)

        while True:
            if shuffle_epoch:
                rand_obj.shuffle(blocks)

            for batch_cnt in range(batches_per_epoch):
                for block_cnt in range(batch_size):
                    block_ind = batch_cnt * batch_size + block_cnt

                    im, seg = blocks[block_ind]
                    if im.ndim == 3:
                        im = im[..., np.newaxis]

                    seg = self.__format_seg_helper__(seg, tissues, include_background)
                    assert im.shape == img_size, "Input shape mismatch. Expected %s, got %s" % (
                        img_size,
                        im.shape,
                    )
                    assert seg.shape == mask_size, "Ouput shape mismatch. Expected %s, got %s" % (
                        mask_size,
                        seg.shape,
                    )

                    x[block_cnt, ...] = im
                    y[block_cnt, ...] = seg

                yield (x, y)

    def __format_seg_helper__(self, seg, tissues, include_background):
        seg = seg[..., np.newaxis]
        seg = np.transpose(seg, [0, 1, 2, 4, 3])

        # support multi class
        if len(tissues) > 1:
            seg_tissues = self._compress_multi_class_mask(seg, tissues)
        else:
            seg_tissues = seg[..., 0, tissues]
        seg_total = seg_tissues

        # if considering background, add class
        # background should mark every other pixel that is not already
        # accounted for in segmentation
        if include_background:
            seg_total = self._add_background_labels(seg_tissues)

        return seg_total

    def _compress_multi_class_mask(self, seg, tissues):
        o_seg = []

        for t_inds in tissues:
            c_seg = seg[..., 0, t_inds]
            if c_seg.ndim == 3:
                c_seg = np.sum(c_seg, axis=-1)
            o_seg.append(c_seg)

        return np.stack(o_seg, axis=-1)

    def summary(self):
        config = self.config

        if config.STATE == "training":
            self.__state_summary(GeneratorState.TRAINING)
            self.__state_summary(GeneratorState.VALIDATION)

            logger.info("INFO: Image size: %s" % (self.config.IMG_SIZE,))
            logger.info("INFO: Image types included in training: %s" % (self.config.FILE_TYPES,))
        else:  # config in Testing state
            self.__state_summary(GeneratorState.TESTING)
            if not config.USE_CROSS_VALIDATION:
                logger.info("Test path: %s" % config.TEST_PATH)

    def __state_summary(self, state: GeneratorState):
        scan_to_blocks, batches_per_epoch, _ = self.cached_data(state)
        volume_ids = scan_to_blocks.keys()
        num_volumes = len(volume_ids)
        num_blocks = self.__get_num_blocks__(scan_to_blocks)

        pids = [self.fname_parser.get_pid_from_volume_id(vol_id) for vol_id in volume_ids]
        num_pids = len(pids)

        base_info = self._img_generator_base_info(state)
        batch_size = base_info["batch_size"]

        logger.info(
            "INFO: %s size: %d blocks (%d volumes), "
            "batch size: %d, # subjects: %d"
            % (state.name, num_blocks, num_volumes, batch_size, num_pids)
        )

    def num_steps(self):
        _, train_batches_per_epoch, _ = self.cached_data(GeneratorState.TRAINING)
        _, valid_batches_per_epoch, _ = self.cached_data(GeneratorState.VALIDATION)

        return train_batches_per_epoch, valid_batches_per_epoch

    def num_examples(self, state: GeneratorState):
        scan_to_blocks, batches_per_epoch, _ = self.cached_data(state)
        num_blocks = self.__get_num_blocks__(scan_to_blocks)
        return num_blocks


class CTGenerator(OAIGenerator):
    """Generator for abdominalCT.

    Includes windowing data as a preprocessing step.
    """

    SUPPORTED_TAGS = ["abct", "abCT"]
    __EXPECTED_IMG_SIZE_DIMS__ = 3

    def __init__(self, config: Config, windows=None):
        if windows and config.num_neighboring_slices() != len(windows):
            raise ValueError("Expected {} windows".format(config.num_neighboring_slices()))
        self.windows = windows
        super().__init__(config)

    def _load_inputs(self, data_path: str, file: str):
        im, seg = self._load_inputs_basic(data_path, file)
        im = self._preprocess(im, self.windows[0] if self.windows else None)
        return im, seg

    def _load_inputs_basic(self, data_path: str, file: str):
        im_path = "%s/%s.im" % (data_path, file)
        with h5py.File(im_path, "r") as f:
            im = f["data"][:]
            if len(im.shape) == 2:
                im = im[..., np.newaxis]

        seg_path = "%s/%s.seg" % (data_path, file)
        with h5py.File(seg_path, "r") as f:
            seg = f["data"][:].astype("float32")
        seg = np.expand_dims(seg, axis=2)  # legacy purposes

        assert len(im.shape) == 3
        assert len(seg.shape) == 4 and seg.shape[-2] == 1

        return im, seg

    def _preprocess(self, im, window):
        # Apply windowing.
        if window:
            im = np.clip(im, window[0], window[1])

        # Preprocess by max normalizing.
        im -= np.min(im)
        im /= np.max(im)

        return im

    def _load_neighboring_slices(self, num_slices, filepath, max_slice):
        """Stacks 2D CT slices from single patient clipped at different window levels.  # noqa

        Overloads traditional 2.5D networks that look at neighboring slices.
        """
        data_path, filename = (os.path.dirname(filepath), os.path.basename(filepath))
        im, seg = self._load_inputs_basic(data_path, filename)
        h, w = im.shape[:2]

        ims = []
        for window in self.windows:
            ims.append(np.squeeze(self._preprocess(np.copy(im), window)))
        im = np.stack(ims, axis=-1)

        assert im.shape == (h, w, self.config.num_neighboring_slices())
        return im, seg

    def img_generator_test(self, model=None):
        config = self.config
        img_size = config.IMG_SIZE
        tissues = config.CATEGORIES
        include_background = config.INCLUDE_BACKGROUND
        num_neighboring_slices = config.num_neighboring_slices()

        base_info = self._img_generator_base_info(GeneratorState.TESTING)
        batch_size = base_info["batch_size"]

        if len(img_size) != self.__EXPECTED_IMG_SIZE_DIMS__:
            raise ValueError("Image size must be {}D".format(self.__EXPECTED_IMG_SIZE_DIMS__))

        files, batches_per_epoch, _ = self._calc_generator_info(GeneratorState.TESTING)
        files = self.sort_files(files)
        scan_id_to_files = self._map_files_to_scan_id(files)
        scan_ids = sorted(scan_id_to_files.keys())

        for scan_id in list(scan_ids):
            scan_id_files = scan_id_to_files[scan_id]
            assert len(scan_id_files) == 1, "Expect 1 slice per scan"
            filepath = scan_id_files[0]

            im, seg_total = self._load_input_helper(
                filepath=filepath,
                tissues=tissues,
                num_neighboring_slices=num_neighboring_slices,
                max_slice_num=1,
                include_background=include_background,
            )
            x = im[np.newaxis, ...]
            y = seg_total[np.newaxis, ...]
            x_orig, _ = self._load_inputs_basic(
                os.path.dirname(filepath), os.path.basename(filepath)
            )

            recon = None
            time_elapsed = None
            if model:
                start_time = time.time()
                recon = model.predict(x, batch_size=batch_size)
                time_elapsed = time.time() - start_time
                x, y, recon = self._reformat_testing_scans((x, y, recon))
            else:
                x, y = self._reformat_testing_scans((x, y))

            yield (x_orig, x, y, recon, scan_id, time_elapsed)



import ast
from typing import Sequence


def convert_data_type(var_string, original):
    """
    Convert string to relevant data type
    :param var_string: variable as a string (e.g.: '[0]', '1', '2.0', 'hellow')
    :param original: original variable
    :return: string converted to data_type
    """
    if isinstance(original, str):
        return var_string

    if isinstance(original, float):
        return float(var_string)

    if isinstance(original, int):
        return int(var_string)

    if isinstance(original, bool):
        return ast.literal_eval(var_string)

    if isinstance(original, Sequence) or isinstance(original, type(None)):
        return ast.literal_eval(var_string)



"""Test built-in metrics.

Most metrics have 2 built-in methods:

    1. `test_consistency`: Test that output of metric is consistent with
        existing implementations (typically from `medpy`)
    2. `test_multiclass`: If the metric supports multi-class, it should have
        a argument named `category_dim` in its callable arguments. The
        results should also be the same as processing each class one by one.

While many of these methods are repeated across classes, we choose not to use
inheritance. This is because each unittest should encapsulate functionality
specific to the class. Instead, we provide helper methods that handle
overlapping functionality:

    * `_binary_arr`: Creates a random binary array of shape `shape`
    * `_is_consistent_medpy`: Helper method for `test_consistency` with
      metrics from the `medpy` library.
    * `_test_multiclass`: Helper method to verify that the multi-class
       implementation works as expected.
"""

import inspect
import unittest

import numpy as np
from medpy.metric import assd, dc, precision, recall

from medsegpy.evaluation.metrics import ASSD, CV, DSC, VOE, Metric, Precision, Recall

_SHAPE = (300, 300, 40)
_NUM_CLASSES = 4
_MULTI_CLASS_SHAPE = _SHAPE + (_NUM_CLASSES,)
_CATEGORY_DIM = -1


def _binary_arr(shape):
    """Make binary array.

    Args:
        shape: Shape of array to make.

    Returns:
        ndarray: A binary array.
    """
    return (np.random.random_sample(shape) > 0.5).astype(np.uint8)


def _is_consistent_medpy(metric: Metric, func, supports_multiclass: bool):
    """Test consistency between metric and medpy implementation.

    Args:
        metric (Metric): A metric
        func (function): MedPy equivalent of metric. Should accept binary input
            in order `result`, `reference`. If you would like to load other
            variables in, provide a function with partial arguments.
        supports_multiclass (bool): If `metric` is expected to support
            multiclass implementation. If `True`, check for duck-typing
            of `category_dim` argument.
    """
    # Single class
    y_true = _binary_arr(_SHAPE)
    y_pred = _binary_arr(_SHAPE)
    args = {"y_pred": y_pred, "y_true": y_true}
    assert np.allclose(func(y_pred, y_true), metric(**args))

    # Multiple classes
    num_classes = _NUM_CLASSES
    y_true = _binary_arr(_MULTI_CLASS_SHAPE)
    y_pred = _binary_arr(_MULTI_CLASS_SHAPE)

    if supports_multiclass:
        category_dim = _CATEGORY_DIM
        args = {"y_pred": y_pred, "y_true": y_true}
        metrics_vals = metric(category_dim=category_dim, **args)
    else:
        metrics_vals = np.asarray(
            [metric(y_pred[..., c], y_true[..., c]) for c in range(num_classes)]
        )

    base_vals = np.asarray(
        [func(result=y_pred[..., c], reference=y_true[..., c]) for c in range(num_classes)]
    )

    assert np.allclose(base_vals, metrics_vals)


def _test_multiclass(metric: Metric):
    """Test multi-class functionality in :class:`Metric`.

    Verify the following:
        1. `metric(...)` is duck-typed with `category_dim` argument
        2. `category_dim` defaults to `None`
        3. `category_dim` can be any dimension

    Args:
        metric:

    Returns:

    """
    arg_names = inspect.getfullargspec(metric).args
    assert (
        "category_dim" in arg_names
    ), "Metrics supporting multiple categories must have 'category_dim' arg"

    idx = arg_names.index("category_dim") - len(arg_names)
    defaults = inspect.getfullargspec(metric).defaults
    assert defaults[idx] is None, "Default value for 'category_dim' should be `None`."

    # Standard category_dim=-1
    shape = (100, 100, 100, 4)
    category_dim = -1
    num_classes = shape[category_dim]
    y_true = _binary_arr(shape)
    y_pred = _binary_arr(shape)

    args = {"y_pred": y_pred, "y_true": y_true}
    metrics_vals = metric(category_dim=category_dim, **args)
    base_vals = np.asarray([metric(y_pred[..., c], y_true[..., c]) for c in range(num_classes)])
    assert np.allclose(base_vals, metrics_vals), "category_dim={}".format(category_dim)

    # Standard category_dim=1
    shape = (100, 6, 100, 100)
    category_dim = 1
    num_classes = shape[category_dim]
    y_true = _binary_arr(shape)
    y_pred = _binary_arr(shape)

    args = {"y_pred": y_pred, "y_true": y_true}
    metrics_vals = metric(category_dim=category_dim, **args)
    base_vals = np.asarray(
        [metric(y_pred[:, c, ...], y_true[:, c, ...]) for c in range(num_classes)]
    )
    assert np.allclose(base_vals, metrics_vals), "category_dim={}".format(category_dim)


class TestDSC(unittest.TestCase):
    def test_consistency(self):
        _is_consistent_medpy(DSC(), dc, True)

    def test_multiclass(self):
        _test_multiclass(DSC())


class TestVOE(unittest.TestCase):
    @staticmethod
    def _expected_voe(result, reference):
        y_true = reference.flatten()
        y_pred = result.flatten()

        y_true_bool = np.asarray(y_true, dtype=np.bool)
        y_pred_bool = np.asarray(y_pred, dtype=np.bool)
        TP = np.sum(y_true_bool * y_pred_bool, axis=-1)
        FP = np.sum(~y_true_bool * y_pred_bool, axis=-1)
        FN = np.sum(y_true_bool * ~y_pred_bool, axis=-1)

        voe = 1 - (TP) / (TP + FP + FN)

        return voe

    def test_consistency(self):
        _is_consistent_medpy(VOE(), self._expected_voe, True)

    def test_multiclass(self):
        _test_multiclass(VOE())


class TestCV(unittest.TestCase):
    @staticmethod
    def _expected_cv(result, reference):
        y_true = np.squeeze(reference)
        y_pred = np.squeeze(result)

        cv = np.std([np.sum(y_true), np.sum(y_pred)]) / np.mean([np.sum(y_true), np.sum(y_pred)])
        return cv

    def test_consistency(self):
        _is_consistent_medpy(CV(), self._expected_cv, True)

    def test_multiclass(self):
        _test_multiclass(CV())


class TestASSD(unittest.TestCase):
    def test_consistency(self):
        _is_consistent_medpy(ASSD(), assd, False)


class TestPrecision(unittest.TestCase):
    def test_consistency(self):
        _is_consistent_medpy(Precision(), precision, True)

    def test_multiclass(self):
        _test_multiclass(Precision())


class TestRecall(unittest.TestCase):
    def test_consistency(self):
        _is_consistent_medpy(Recall(), recall, True)

    def test_multiclass(self):
        _test_multiclass(Recall())


if "__name__" == "__main__":
    unittest.main()



import os
import pickle
import re
from abc import ABC
from typing import Any

import h5py
from fvcore.common.file_io import PathHandler, PathManager

from medsegpy.utils.cluster import Cluster


def load_h5(file_path):
    """Load data in h5df format.

    Args:
        file_path (str): Path to h5 file

    Returns:
        dict: dictionary representation of h5df data.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError("%s does not exist" % file_path)

    data = dict()
    with h5py.File(file_path, "r") as f:
        for key in f.keys():
            data[key] = f.get(key).value

    return data


def save_optimizer(optimizer, dirpath: str):
    """Serialize a model and add the config of the optimizer.

    Args:
        optimizer (keras.Optimzer): a Keras optimizer
        dirpath (str): Path to directory
    """
    if optimizer is None:
        return

    config = dict()
    config["optimizer"] = optimizer.get_config()

    filepath = os.path.join(dirpath, "optimizer.dat")
    # Save optimizer state
    save_pik(config, filepath)


def load_optimizer(dirpath: str):
    """Return model and optimizer in previous state."""
    from keras import optimizers

    filepath = os.path.join(dirpath, "optimizer.dat")
    model_dict = load_pik(filepath)
    optimizer_params = {k: v for k, v in model_dict.get("optimizer").items()}
    optimizer = optimizers.get(optimizer_params)

    return optimizer


def save_pik(data, filepath):
    """Save data using pickle."""
    with open(filepath, "wb") as f:
        pickle.dump(data, f)


def load_pik(filepath):
    """
    Load data using pickle
    :param filepath: filepath to load from
    :return: data saved using save_pik
    """
    with open(filepath, "rb") as f:
        return pickle.load(f)


def format_exp_version(dir_path, new_version=True, mkdirs=False, force=False):
    """Adds experiment version to the directory path. Returns local path.

    If `os.path.basename(dir_path)` starts with 'version', assume the version
    has already been formatted.

    Args:
        dir_path (str): The directory path corresponding to the version.
        force (bool, optional): If `True` force adds version even if 'version'
            is part of basename.

    Returns:
        str: The formatted dirpath
    """
    dir_path = PathManager.get_local_path(dir_path)
    if not os.path.isdir(dir_path):
        return os.path.join(dir_path, "version_001")
    if not force and re.match("^version_[0-9]*", os.path.basename(dir_path)):
        return dir_path
    version_dir, version_num = _find_latest_version_dir(dir_path)
    if new_version:
        version_num += 1
        version_dir = f"version_{version_num:03d}"
    version_dirpath = os.path.join(dir_path, version_dir)
    if mkdirs:
        PathManager.mkdirs(version_dirpath)
    return version_dirpath


def _find_latest_version_dir(dir_path):
    version_dirs = [
        (x, int(x.split("_")[1])) for x in os.listdir(dir_path) if re.match("^version_[0-9]*", x)
    ]
    if len(version_dirs) == 0:
        version_dir, version_num = None, 0
    else:
        version_dirs = sorted(version_dirs, key=lambda x: x[1])
        version_dir, version_num = version_dirs[-1]
    return version_dir, version_num


class GeneralPathHandler(PathHandler, ABC):  # noqa: B024
    PREFIX = ""

    def _get_supported_prefixes(self):
        return [self.PREFIX]

    def _get_local_path(self, path: str, **kwargs: Any):
        name = path[len(self.PREFIX) :]
        return os.path.join(Cluster.working_cluster().results_dir, self._project_name(), name)

    def _open(self, path, mode="r", **kwargs):
        return PathManager.open(self._get_local_path(path), mode, **kwargs)

    def _mkdirs(self, path: str, **kwargs: Any):
        os.makedirs(self._get_local_path(path), exist_ok=True)

    def _project_name(self):
        return self.PREFIX[:-3]


class GeneralResultsPathHandler(GeneralPathHandler):
    PREFIX = "results://"

    def _project_name(self):
        return ""


class TechConsiderationsHandler(GeneralPathHandler):
    PREFIX = "tcv3://"

    def _project_name(self):
        return "tech-considerations"


class Context_Self_Supervision_Handler(GeneralPathHandler):
    PREFIX = "css://"

    def _project_name(self):
        return "context-self-supervision"


PathManager.register_handler(GeneralResultsPathHandler())
PathManager.register_handler(TechConsiderationsHandler())
PathManager.register_handler(Context_Self_Supervision_Handler())



"""OAI iMorphics Dataset"""
import logging
import os
import re

import h5py

from medsegpy.data.catalog import DatasetCatalog, MetadataCatalog
from medsegpy.utils.cluster import Cluster

logger = logging.getLogger(__name__)

_DATA_CATALOG = {}
_TEST_SET_METADATA_PIK = ""

_DATA_CATALOG = {
    "oai_2d_train": "oai_data/h5_files_2d/train",
    "oai_2d_val": "oai_data/h5_files_2d/valid",
    "oai_2d_test": "oai_data/h5_files_2d/test",
    "oai_2d_whitened_train": "oai_data/h5_files_whitened_2d/train",
    "oai_2d_whitened_val": "oai_data/h5_files_whitened_2d/valid",
    "oai_2d_whitened_test": "oai_data/h5_files_whitened_2d/test",
    "oai_3d_train": "oai_data/h5_files_3d/train",
    "oai_3d_val": "oai_data/h5_files_3d/val",
    "oai_3d_test": "oai_data/h5_files_3d/test",
    "oai_3d_whitened_train": "oai_data/h5_files_whitened_3d/train",
    "oai_3d_whitened_val": "oai_data/h5_files_whitened_3d/val",
    "oai_3d_whitened_test": "oai_data/h5_files_whitened_3d/test",
    "oai_3d_sf_whitened_train": "oai_data/h5_files_whitened_3d/train.h5",
    "oai_3d_sf_whitened_val": "oai_data/h5_files_whitened_3d/val.h5",
    "oai_3d_sf_whitened_test": "oai_data/h5_files_whitened_3d/test.h5",
}


OAI_CATEGORIES = [
    {"color": [220, 20, 60], "id": 0, "name": "femoral cartilage", "abbrev": "fc"},
    {"color": [119, 11, 32], "id": (1, 2), "name": "tibial cartilage", "abbrev": "tc"},
    {"color": [0, 0, 142], "id": 3, "name": "patellar cartilage", "abbrev": "pc"},
    {"color": [0, 0, 230], "id": (4, 5), "name": "meniscus", "abbrev": "men"},
]


def load_oai_2d_from_dir(scan_root, dataset_name=None):
    # sample scan name: "9311328_V01-Aug04_072.im"
    # format: %7d_V%02d-Aug%02d_%03d
    FNAME_REGEX = "([\d]+)_V([\d]+)-Aug([\d]+)_([\d]+)"

    files = sorted(os.listdir(scan_root))
    filepaths = [os.path.join(scan_root, f) for f in files if f.endswith(".im")]

    dataset_dicts = []
    for fp in filepaths:
        _, pid, time_point, _, slice_id, _ = tuple(re.split(FNAME_REGEX, fp))
        pid = int(pid)
        time_point = int(time_point)
        dataset_dicts.append(
            {
                "file_name": fp,
                "sem_seg_file": "{}.seg".format(os.path.splitext(fp)[0]),
                "scan_id": "{:07d}_V{:02d}".format(pid, time_point),
                "subject_id": pid,
                "time_point": time_point,
                "slice_id": int(slice_id),
                "scan_num_slices": 160,
            }
        )

    num_scans = len({d["scan_id"] for d in dataset_dicts})
    num_subjects = len({d["subject_id"] for d in dataset_dicts})
    if dataset_name:
        logger.info("Loaded {} from {}".format(dataset_name, scan_root))
    logger.info(
        "Loaded {} scans from {} subjects ({} slices)".format(
            num_scans, num_subjects, len(dataset_dicts)
        )
    )

    return dataset_dicts


def load_oai_3d_from_dir(scan_root, dataset_name=None):
    # sample scan name: "9311328_V01-Aug04_072.h5"
    FNAME_REGEX = "([\d]+)_V([\d]+)"

    files = sorted(os.listdir(scan_root))
    filepaths = [os.path.join(scan_root, f) for f in files]
    dataset_dicts = []
    for fp in filepaths:
        _, pid, time_point, _ = tuple(re.split(FNAME_REGEX, fp))
        pid = int(pid)
        time_point = int(time_point)
        dataset_dicts.append(
            {
                "file_name": fp,
                "sem_seg_file": fp,
                "scan_id": "{:07d}_V{:02d}".format(pid, time_point),
                "subject_id": pid,
                "time_point": time_point,
                "image_size": (384, 384, 160),
            }
        )

    num_scans = len(dataset_dicts)
    num_subjects = len({d["subject_id"] for d in dataset_dicts})
    if dataset_name:
        logger.info("Loaded {} from {}".format(dataset_name, scan_root))
    logger.info("Loaded {} scans from {} subjects".format(num_scans, num_subjects))

    return dataset_dicts


def load_oai_3d_sf_from_dir(scan_root, dataset_name=None):
    """
    Expected file structure:
          keys=> image_files (e.g. '0000001_V00');
          subkeys=> image type (e.g. ['seg', 'volume']);
    """
    FNAME_REGEX = "([\d]+)_V([\d]+)"
    f = h5py.File(scan_root, "r")
    keys = list(f.keys())
    f.close()
    dataset_dicts = []
    for key in keys:
        _, pid, time_point, _ = tuple(re.split(FNAME_REGEX, key))
        pid = int(pid)
        time_point = int(time_point)
        dataset_dicts.append(
            {
                "file_name": key,
                "sem_seg_file": key,
                "scan_id": "{:07d}_V{:02d}".format(pid, time_point),
                "subject_id": pid,
                "time_point": time_point,
                "image_size": (384, 384, 160),
                "singlefile_path": scan_root,
            }
        )

    num_scans = len(dataset_dicts)
    num_subjects = len({d["subject_id"] for d in dataset_dicts})
    if dataset_name:
        logger.info("Loaded {} from {}".format(dataset_name, scan_root))
    logger.info("Loaded {} scans from {} subjects".format(num_scans, num_subjects))

    return dataset_dicts


def register_oai(name, scan_root):
    load_func = None
    if name.startswith("oai_2d"):
        load_func = load_oai_2d_from_dir
    elif name.startswith("oai_3d_sf"):
        load_func = load_oai_3d_sf_from_dir
    else:
        load_func = load_oai_3d_from_dir

    DatasetCatalog.register(name, lambda: load_func(scan_root, name))

    # 2. Optionally, add metadata about this dataset,
    # since they might be useful in evaluation, visualization or logging

    MetadataCatalog.get(name).set(
        scan_root=scan_root,
        spacing=(0.3125, 0.3125, 0.7),
        test_set_metadata_pik=_TEST_SET_METADATA_PIK,
        category_ids=[x["id"] for x in OAI_CATEGORIES],
        category_abbreviations=[x["abbrev"] for x in OAI_CATEGORIES],
        categories=[x["name"] for x in OAI_CATEGORIES],
        category_colors=[x["color"] for x in OAI_CATEGORIES],
        category_id_to_contiguous_id={x["id"]: idx for idx, x in enumerate(OAI_CATEGORIES)},
        evaluator_type="SemSegEvaluator",
    )


def register_all_oai():
    for dataset_name, scan_root in _DATA_CATALOG.items():
        if not os.path.isabs(scan_root):
            scan_root = os.path.join(Cluster.working_cluster().data_dir, scan_root)
        register_oai(dataset_name, scan_root)



import logging
import os

import keras.backend as K
import numpy as np
from keras.initializers import he_normal
from keras.layers import Activation, Add
from keras.layers import BatchNormalization as BN
from keras.layers import (
    Concatenate,
    Conv2D,
    Conv2DTranspose,
    Dense,
    Dropout,
    GlobalAveragePooling2D,
    Input,
    MaxPooling2D,
    Permute,
    Reshape,
    multiply,
)
from keras.models import Model
from keras.utils import plot_model
from keras.utils.layer_utils import get_source_inputs

logger = logging.getLogger(__name__)

# List of tissues that can be segmented
FEMORAL_CARTILAGE_STR = "fc"
MENISCUS_STR = "men"
PATELLAR_CARTILAGE_STR = "pc"
TIBIAL_CARTILAGE_STR = "tc"

# Absolute directory where this file lives
__ABS_DIR__ = os.path.dirname(os.path.abspath(__file__))

WEIGHTS_DICT = {
    FEMORAL_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_fc_weights--0.8968.h5"),
    MENISCUS_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_men_weights--0.7692.h5"),
    PATELLAR_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_pc_weights--0.6206.h5"),
    TIBIAL_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_tc_weights--0.8625.h5"),
}

# Input size that is expected
# All inputs must be at least this size
DEFAULT_INPUT_SIZE = (288, 288, 1)


def squeeze_excitation_block(xi, base_name: str, seed=None, ratio=8):
    """Adapted from https://github.com/titu1994/keras-squeeze-excite-network"""
    x = xi
    channel_axis = 1 if K.image_data_format() == "channels_first" else -1
    filters = x._keras_shape[channel_axis]
    se_shape = (1, 1, filters)

    se = GlobalAveragePooling2D(name="%s_se_glob_avg_pool" % base_name)(x)
    se = Reshape(se_shape, name="%s_se_reshape" % base_name)(se)
    se = Dense(
        filters // ratio,
        name="%s_se_fc1" % base_name,
        kernel_initializer=he_normal(seed),
        use_bias=False,
    )(se)
    se = Activation("relu", name="%s_se_relu" % base_name)(se)
    se = Dense(filters, use_bias=False)(se)
    se = Activation("sigmoid", name="%s_se_sigmoid" % base_name)(se)

    if K.image_data_format() == "channels_first":
        se = Permute((3, 1, 2))(se)

    x = multiply([x, se], name="%s_se_scale" % base_name)
    return x


def res_block(
    xi,
    base_name: str,
    nfeatures: int,
    nlayers: int = 2,
    kernel_size=(3, 3),
    seed=None,
    layer_order=("conv", "relu", "bn"),
    dropout_rate=0.0,
    use_squeeze_excitation=False,
    squeeze_excitation_ratio=8,
):
    x = xi

    layers = []
    for l in range(nlayers):  # noqa: E741
        layers_dict = {
            "conv": Conv2D(
                nfeatures,
                kernel_size=kernel_size,
                padding="same",
                kernel_initializer=he_normal(seed=seed),
                name="%s_conv_%d" % (base_name, l + 1),
            ),
            "bn": BN(axis=-1, momentum=0.95, epsilon=0.001, name="%s_bn_%d" % (base_name, l + 1)),
            "relu": Activation("relu", name="%s_relu_%d" % (base_name, l + 1)),
            "dropout": Dropout(rate=dropout_rate, name="%s_dropout_%d" % (base_name, l + 1)),
        }

        for layer in layer_order:
            x = layers_dict[layer](x)
            layers.append(x)

    if use_squeeze_excitation:
        x = squeeze_excitation_block(
            x, base_name=base_name, seed=seed, ratio=squeeze_excitation_ratio
        )

    x = Add()([x, xi])

    return x


def residual_unet_2d(
    input_size=DEFAULT_INPUT_SIZE,
    input_tensor=None,
    output_mode=None,
    num_filters=None,
    depth=6,
    dropout_rate=0.0,
    layer_order=("relu", "bn", "dropout", "conv"),
    use_squeeze_excitation=False,
    squeeze_excitation_ratio=8,
):
    """Generate Unet 2D model compatible with Keras 2

    :param input_size: tuple of input size - format: (height, width, 1)

    :rtype: Keras model

    :raise ValueError if input_size is not tuple or dimensions of input_size do not match
        (height, width, 1)
    """
    from medsegpy import glob_constants

    logger.info("Initializing unet with seed: %s" % str(glob_constants.SEED))
    SEED = glob_constants.SEED
    if input_tensor is None and (type(input_size) is not tuple or len(input_size) != 3):
        raise ValueError("input_size must be a tuple of size (height, width, 1)")

    if num_filters is None:
        nfeatures = [2**feat * 32 for feat in np.arange(depth)]
    else:
        nfeatures = num_filters
        assert len(nfeatures) == depth

    conv_ptr = []

    # input layer
    inputs = input_tensor if input_tensor is not None else Input(input_size)

    # step down convolutional layers
    pool = inputs
    for depth_cnt in range(depth):
        conv = Conv2D(
            nfeatures[depth_cnt],
            kernel_size=(3, 3),
            padding="same",
            kernel_initializer=he_normal(seed=SEED),
            name="enc_%d_input" % (depth_cnt + 1),
        )(pool)

        conv = res_block(
            conv,
            base_name="enc_%d" % (depth_cnt + 1),
            nfeatures=nfeatures[depth_cnt],
            layer_order=layer_order,
            dropout_rate=dropout_rate,
            seed=SEED,
            use_squeeze_excitation=use_squeeze_excitation,
            squeeze_excitation_ratio=squeeze_excitation_ratio,
        )

        conv = Activation("relu", name="enc_%d_ouput_relu" % (depth_cnt + 1))(conv)
        conv = BN(axis=-1, momentum=0.95, epsilon=0.001, name="enc_%d_ouput_bn" % (depth_cnt + 1))(
            conv
        )

        conv_ptr.append(conv)

        # Only maxpool till penultimate depth
        if depth_cnt < depth - 1:

            # If size of input is odd, only do a 3x3 max pool
            xres = conv.shape.as_list()[1]
            if xres % 2 == 0:
                pooling_size = (2, 2)
            elif xres % 2 == 1:
                pooling_size = (3, 3)

            pool = MaxPooling2D(pool_size=pooling_size, name="enc_maxpool_%d" % (depth_cnt + 1))(
                conv
            )

    # step up convolutional layers
    for depth_cnt in range(depth - 2, -1, -1):

        deconv_shape = conv_ptr[depth_cnt].shape.as_list()
        deconv_shape[0] = None

        # If size of input is odd, then do a 3x3 deconv
        if deconv_shape[1] % 2 == 0:
            unpooling_size = (2, 2)
        elif deconv_shape[1] % 2 == 1:
            unpooling_size = (3, 3)

        up = Concatenate(axis=3)(
            [
                Conv2DTranspose(
                    nfeatures[depth_cnt],
                    (3, 3),
                    padding="same",
                    strides=unpooling_size,
                    kernel_initializer=he_normal(seed=SEED),
                    name="dec_deconv_%d" % (depth_cnt + 1),
                )(conv),
                conv_ptr[depth_cnt],
            ]
        )

        up = Conv2D(
            nfeatures[depth_cnt],
            kernel_size=(3, 3),
            padding="same",
            kernel_initializer=he_normal(seed=SEED),
            name="dec_%d_input" % (depth_cnt + 1),
        )(up)

        conv = res_block(
            up,
            base_name="dec_%d" % (depth_cnt + 1),
            nfeatures=nfeatures[depth_cnt],
            seed=SEED,
            layer_order=layer_order,
            use_squeeze_excitation=use_squeeze_excitation,
            squeeze_excitation_ratio=squeeze_excitation_ratio,
        )

        conv = Activation("relu", name="dec_%d_ouput_relu" % (depth_cnt + 1))(conv)
        conv = BN(axis=-1, momentum=0.95, epsilon=0.001, name="dec_%d_ouput_bn" % (depth_cnt + 1))(
            conv
        )

    if output_mode is not None:
        recon = Conv2D(
            1, (1, 1), padding="same", kernel_initializer=he_normal(seed=SEED), name="1d_class_conv"
        )(conv)
        recon = Activation(output_mode, name="class_%s" % output_mode)(recon)
    else:
        recon = conv

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)

    model = Model(inputs=inputs, outputs=[recon])

    return model


if __name__ == "__main__":
    save_path = "../imgs/res_unet2d_se.png"
    m = residual_unet_2d(use_squeeze_excitation=True)
    plot_model(m, to_file=save_path)
